from ._typing import OnchainTxType
from .utils import HexUtil


class Constant(object):
    MINT_KECCAK = "0x7a53080ba414158be7ec69b987b5fb7d07dee101fe85488f0853ae16239d0bde"
    SWAP_KECCAK = "0xc42079f94a6350d7e6235f29174924f928cc2ac818eb64fed8004e115fbcca67"
    BURN_KECCAK = "0x0c396cd989a39f4459b5fa1aed6a9a8dcdbc45908acfd67e028cd568da98982c"
    COLLECT_KECCAK = "0x70935338e69775456a85ddef226c395fb668b63fa0115f5f20610b388e6ca9c0"


type_dict = {Constant.MINT_KECCAK: OnchainTxType.MINT,
             Constant.SWAP_KECCAK: OnchainTxType.SWAP,
             Constant.BURN_KECCAK: OnchainTxType.BURN,
             Constant.COLLECT_KECCAK: OnchainTxType.COLLECT}


def decode_address_from_topic(topic_str):
    return "0x" + topic_str[26:]


def split_topic(topic: str) -> list[str]:
    spliter = "\n" if "\n" in topic else ","
    topic_list = topic.strip("[]").replace("'", "").replace("\"", "").replace(" ", "").split(spliter)
    return topic_list


def handle_event(topics_str, data_hex):
    # proprocess topics string -> topic list
    # topics_str = topics.values[0]
    sqrtPriceX96 = receipt = amount1 = current_liquidity = current_tick = tick_lower = tick_upper = delta_liquidity = None
    if isinstance(topics_str, str):
        topic_list = split_topic(topics_str)
    else:
        topic_list = topics_str

    # data_hex = data.values[0]

    type_topic = topic_list[0]
    tx_type = type_dict[type_topic]
    no_0x_data = data_hex[2:]
    chunk_size = 64
    chunks = len(no_0x_data)

    if tx_type == OnchainTxType.SWAP:
        sender = decode_address_from_topic(topic_list[1])
        receipt = decode_address_from_topic(topic_list[2])
        split_data = ["0x" + no_0x_data[i:i + chunk_size] for i in range(0, chunks, chunk_size)]
        amount0, amount1, sqrtPriceX96, current_liquidity, current_tick = [HexUtil.to_signed_int(onedata) for onedata in
                                                                           split_data]
    elif tx_type == OnchainTxType.BURN:
        sender = decode_address_from_topic(topic_list[1])
        tick_lower = HexUtil.to_signed_int(topic_list[2])
        tick_upper = HexUtil.to_signed_int(topic_list[3])
        split_data = ["0x" + no_0x_data[i:i + chunk_size] for i in range(0, chunks, chunk_size)]
        delta_liquidity, amount0, amount1 = [HexUtil.to_signed_int(onedata) for onedata in split_data]
        delta_liquidity = -delta_liquidity
    elif tx_type == OnchainTxType.MINT:
        # sender = topic_str_to_address(topic_list[1])
        owner = decode_address_from_topic(topic_list[1])
        tick_lower = HexUtil.to_signed_int(topic_list[2])
        tick_upper = HexUtil.to_signed_int(topic_list[3])
        split_data = ["0x" + no_0x_data[i:i + chunk_size] for i in range(0, chunks, chunk_size)]
        sender = decode_address_from_topic(split_data[0])
        delta_liquidity, amount0, amount1 = [HexUtil.to_signed_int(onedata) for onedata in split_data[1:]]
    elif tx_type == OnchainTxType.COLLECT:
        tick_lower = HexUtil.to_signed_int(topic_list[2])
        tick_upper = HexUtil.to_signed_int(topic_list[3])
        split_data = ["0x" + no_0x_data[i:i + chunk_size] for i in range(0, chunks, chunk_size)]
        sender = decode_address_from_topic(split_data[0])
        amount0, amount1 = [HexUtil.to_signed_int(onedata) for onedata in split_data[1:]]
    else:
        raise ValueError("not support tx type")

    return tx_type, sender, receipt, amount0, amount1, sqrtPriceX96, current_liquidity, current_tick, tick_lower, tick_upper, delta_liquidity



from dataclasses import dataclass
from datetime import datetime, timedelta

from demeter._typing import RowData, DemeterError


def to_minute(time: datetime) -> datetime:
    return datetime(time.year, time.month, time.day, time.hour, time.minute)


class Trigger:
    def __init__(self, do, *args, **kwargs):
        self._do = do if do is not None else self.do_nothing
        self.kwargs = kwargs
        self.args = args

    def when(self, row_data: RowData) -> bool:
        return False

    def do_nothing(self, row_data: RowData, *args, **kwargs):
        pass

    def do(self, row_data: RowData):
        return self._do(row_data, *self.args, **self.kwargs)


class AtTimeTrigger(Trigger):
    def __init__(self, time: datetime, do, *args, **kwargs):
        self._time = to_minute(time)
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return row_data.timestamp == self._time


class AtTimesTrigger(Trigger):
    def __init__(self, time: [datetime], do, *args, **kwargs):
        self._time = [to_minute(t) for t in time]
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._time in row_data.timestamp


@dataclass
class TimeRange:
    start: datetime
    end: datetime


class TimeRangeTrigger(Trigger):
    def __init__(self, time_range: TimeRange, do, *args, **kwargs):
        self._time_range = TimeRange(to_minute(time_range.start), to_minute(time_range.end))
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._time_range.start <= row_data.timestamp < self._time_range.end


class TimeRangesTrigger(Trigger):
    def __init__(self, time_range: [TimeRange], do, *args, **kwargs):
        self._time_range: [TimeRange] = [TimeRange(to_minute(t.start), to_minute(t.end)) for t in time_range]
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: RowData) -> bool:
        for r in self._time_range:
            if r.start <= row_data.timestamp < r.end:
                return True
        return False


def check_time_delta(delta: timedelta):
    if delta.total_seconds() % 60 != 0:
        raise DemeterError("min time span is 1 minute")


class PeriodTrigger(Trigger):
    def __init__(self, time_delta: timedelta, do, trigger_immediately=False, *args, **kwargs):
        self._next_match = None
        self._delta = time_delta
        self._trigger_immediately = trigger_immediately
        check_time_delta(time_delta)
        super().__init__(do, *args, **kwargs)

    def reset(self):
        self._next_match = None

    def when(self, row_data: RowData) -> bool:
        if self._next_match is None:
            self._next_match = row_data.timestamp + self._delta
            return self._trigger_immediately

        if self._next_match == row_data.timestamp:
            self._next_match = self._next_match + self._delta
            return True

        return False


class PeriodsTrigger(Trigger):
    def __init__(self, time_delta: [timedelta], do, trigger_immediately=False, *args, **kwargs):
        self._next_matches = [None for _ in time_delta]
        self._deltas = time_delta
        self._trigger_immediately = trigger_immediately

        for td in time_delta:
            check_time_delta(td)
        super().__init__(do, *args, **kwargs)

    def reset(self):
        self._next_matches = [None for _ in self._delta]

    def when(self, row_data: RowData) -> bool:
        if self._next_matches[0] is None:
            self._next_matches = [row_data.timestamp + d for d in self._delta]
            return self._trigger_immediately

        for i in range(len(self._deltas)):
            if self._next_matches[i] == row_data.timestamp:
                self._next_matches[i] = self._next_matches[i] + self._deltas[i]
                return True

        return False



from decimal import Decimal

import pandas as pd

from .._typing import AccountStatus, EvaluatingIndicator, DECIMAL_ZERO, UnitDecimal, DemeterError

from typing import Union


class Evaluator(object):
    """
    calculate evaluator indicator for strategy.
    """

    def __init__(self, init_status: AccountStatus, data: Union[pd.DataFrame, AccountStatus]):
        self.init_status: AccountStatus = init_status
        self.init_net_value = init_status.base_balance + init_status.quote_balance * init_status.price
        self.end_status = data.iloc[-1]
        self.data: Union[pd.DataFrame, AccountStatus] = data
        if len(data) < 2:
            raise DemeterError("not enough data")
        self.time_span_in_day = len(data.index) * (data.index[1] - data.index[0]).seconds / (60 * 60 * 24)
        self._evaluating_indicator: EvaluatingIndicator = None

    def run(self):
        self._evaluating_indicator = EvaluatingIndicator(DECIMAL_ZERO, DECIMAL_ZERO, DECIMAL_ZERO)
        self._evaluating_indicator.annualized_returns = UnitDecimal(self.get_annualized_returns(), "")
        self._evaluating_indicator.benchmark_returns = UnitDecimal(self.get_benchmark_returns(), "")
        self._evaluating_indicator.max_drawdown = UnitDecimal(Evaluator.get_max_drawdown_fast(self.data.net_value), "")
        return self._evaluating_indicator

    def get_annualized_returns(self):
        """Annualized return rate"""
        return (self.end_status.net_value / self.init_net_value) ** Decimal(365 / self.time_span_in_day) - 1

    def get_benchmark_returns(self):
        """
        Annualized benchmark return rate
        algorithm: swap token balance to 1:1, and hold those position to the end.
        :return:
        """
        base_amount, quote_amount = self.__get_benchmark_asset()
        final_benchmark_capital = base_amount + quote_amount * self.end_status.price
        return (final_benchmark_capital / self.init_net_value) ** Decimal(365 / self.time_span_in_day) - 1

    @staticmethod
    def get_max_drawdown(net_values: pd.Series):

        net_values.index = range(len(net_values.index))  # restruct index to access faster
        max_drawdown = 0
        for index, row in net_values.iteritems():
            current_max = net_values[index:].apply(lambda nv: 1 - nv / row).max()
            if current_max > max_drawdown:
                max_drawdown = current_max
        return max_drawdown

    @staticmethod
    def get_max_drawdown_fast(net_values: pd.Series):
        max_value, idx_h, idx_l = Evaluator._withdraw_with_high_low(net_values.to_list())
        return (net_values.iloc[idx_h] - net_values.iloc[idx_l]) / net_values.iloc[idx_h]

    @staticmethod
    def _withdraw_with_high_low(arr: list):
        """
        from : https://blog.csdn.net/Spade_/article/details/112341428
        """

        # 传入一个数组，返回最大回撤和对应的最高点索引、最低点索引
        _dp = 0  # 使用 _dp 表示 i 点的最大回撤
        i_high = 0  # 遍历时，0 ~ i - 1 中最高的点的索引，注意是索引

        # 全局最大回撤和对应的最高点和最低点的索引，注意是索引
        g_withdraw, g_high, g_low = float('-inf'), -1, -1

        for i in range(1, len(arr)):
            if arr[i_high] < arr[i - 1]:  # 若 0 ~ i - 1 中最高的点小于当前点
                i_high = i - 1  # 0 ~ i - 1 中最高的点的索引

            _dp = arr[i_high] - arr[i]  # _dp 表示 i 点的最大回撤
            if _dp > g_withdraw:  # 找到新的最大回撤，更新三个值
                g_withdraw = _dp
                g_high = i_high
                g_low = i

        return g_withdraw, g_high, g_low

    def __get_benchmark_asset(self):
        base_amount = self.init_net_value / 2
        quote_amount = (self.init_net_value - base_amount) / self.init_status.price
        return base_amount, quote_amount

    @property
    def evaluating_indicator(self):
        return self._evaluating_indicator



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import NamedTuple

from .._typing import TokenInfo, DemeterError


class PoolBaseInfo(object):
    """
    pool information, corresponding with definition in pool contract.

    :param token0: First token in  pool contract.
    :type token0:  TokenInfo
    :param token1: Second token in  pool contract.
    :type token1: TokenInfo
    :param fee: fee rate of this pool, should be among [0.05, 0.3, 1]
    :type fee: float
    :param base_token: which token will be considered as base token. eg: to a token pair of USDT/BTC, if you want price unit to be like 10000 usdt/btc, you should set usdt as base token, otherwise if price unit is 0.00001 btc/usdt, you should set btc as base token
    :type base_token: TokenInfo
    """

    def __init__(self, token0: TokenInfo, token1: TokenInfo, fee: float, base_token):
        self.token0 = token0
        self.token1 = token1
        self.is_token0_base = (base_token == token0)
        self.base_token = base_token
        fee = str(fee)  # keep precision
        match fee:
            case "0.05":
                self.tickSpacing = 10
            case "0.3":
                self.tickSpacing = 60
            case "1":
                self.tickSpacing = 200
            case _:
                raise DemeterError("fee should be 0.05 or 0.3 or 1")
        self.fee = Decimal(fee) * Decimal(10000)
        self.fee_rate = Decimal(fee) / Decimal(100)

    def __str__(self):
        """
        get string
        :return:
        :rtype:
        """
        return "PoolBaseInfo(Token0: {},".format(self.token0) + \
               "Token1: {},".format(self.token1) + \
               "fee: {},".format(self.fee_rate * Decimal(100)) + \
               "base token: {})".format(self.token0.name if self.is_token0_base else self.token1.name)


class BrokerAsset(object):
    """
    Wallet of broker, manage balance of an asset.
    It will prevent excess usage on asset.
    """

    def __init__(self, token: TokenInfo, init_amount=Decimal(0)):
        self.token_info = token
        self.name = token.name
        self.decimal = token.decimal
        self.balance = init_amount

    def __str__(self):
        return self.name

    def add(self, amount=Decimal(0)):
        """
        add amount to balance
        :param amount: amount to add
        :type amount: Decimal
        :return: entity itself
        :rtype: BrokerAsset
        """
        self.balance += amount
        return self

    def sub(self, amount=Decimal(0), allow_negative_balance=False):
        """
        subtract amount from balance. if balance is not enough, an error will be raised.

        :param amount: amount to subtract
        :type amount: Decimal
        :param allow_negative_balance: allow balance is negative
        :type allow_negative_balance: bool
        :return:
        :rtype:
        """
        base = self.balance if self.balance != Decimal(0) else Decimal(amount)

        if base == Decimal(0):  # amount and balance is both 0
            return self
        if allow_negative_balance:
            self.balance -= amount
        else:
            # if difference between amount and balance is below 0.01%, will deduct all the balance
            # That's because, the amount calculated by v3_core, has some acceptable error.
            if abs((self.balance - amount) / base) < 0.00001:
                self.balance = Decimal(0)
            elif self.balance - amount < Decimal(0):
                raise DemeterError(
                    f"insufficient balance, balance is {self.balance}{self.name}, but sub amount is {amount}{self.name}")
            else:
                self.balance -= amount

        return self

    def amount_in_wei(self):
        return self.balance * Decimal(10 ** self.decimal)


@dataclass
class Position(object):
    """
    variables for position
    """

    pending_amount0: Decimal
    pending_amount1: Decimal
    liquidity: int


class PoolStatus(NamedTuple):
    """
    current status of a pool, actuators can notify current status to broker by filling this entity
    """
    timestamp: datetime
    current_tick: int
    current_liquidity: Decimal
    in_amount0: Decimal
    in_amount1: Decimal
    price: Decimal



import json
from decimal import Decimal
from types import SimpleNamespace

OUTPUT_WIDTH = 30


def object_to_decimal(num):
    return Decimal(str(num)) if (isinstance(num, float) or type(num) == int) else num


def dict_to_object(dict_entity):
    return json.loads(json.dumps(dict_entity), object_hook=lambda d: SimpleNamespace(**d))


def float_param_formatter(func):
    def wrapper_func(*args, **kwargs):
        new_args = ()
        for arg in args:
            new_args += (object_to_decimal(arg),)
        for k, v in kwargs.items():
            kwargs[k] = object_to_decimal(v)
        return func(*new_args, **kwargs)

    return wrapper_func


def get_formatted_str(values: dict) -> str:
    str_array = []
    for k, v in values.items():
        value_str = str(v)
        str_array.append(f"""\033[1;34m{k:<10}:\033[0m {value_str:<30} """)
    return "".join(str_array)


def get_enum_by_name(me, name):
    for e in me:
        if e.name.lower() == name.lower():
            return e
    raise RuntimeError(f"cannot found {name} in {me}, allow value is " + str([x.name for x in me]))



from datetime import timedelta

import numpy as np
import pandas as pd
from pandas._typing import TimedeltaConvertibleTypes, Axis

from .common import get_real_n


def simple_moving_average(data: pd.Series | pd.DataFrame,
                          window: timedelta = timedelta(hours=5),
                          min_periods: int | None = None,
                          center: bool = False,
                          win_type: str | None = None,
                          on: str | None = None,
                          axis: Axis = 0,
                          closed: str | None = None,
                          method: str = "single",
                          ) -> pd.Series:
    """
    calculate simple moving average, Note: window is based on time span

    docs for other params, see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.rolling.html

    :param data: data
    :type data: Series
    :param window: window width
    :type window: timedelta
    :return: simple moving average data
    :rtype: Series

    """

    return data.rolling(window=get_real_n(data, window),
                        min_periods=min_periods,
                        center=center,
                        win_type=win_type,
                        on=on,
                        axis=axis,
                        closed=closed,
                        method=method,
                        ).mean()


def exponential_moving_average(data: pd.Series | pd.DataFrame,
                               com: float | None = None,
                               span: float | None = None,
                               halflife: float | TimedeltaConvertibleTypes | None = None,
                               alpha: float | None = None,
                               min_periods: int | None = 0,
                               adjust: bool = True,
                               ignore_na: bool = False,
                               axis: Axis = 0,
                               times: str | np.ndarray | pd.DataFrame | pd.Series | None = None,
                               method: str = "single",
                               ):
    """
    calculate exponential moving average, just a shortcut for pandas.evm().mean()

    docs for params, see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.ewm.html
    """
    return data.ewm(com=com,
                    span=span,
                    halflife=halflife,
                    alpha=alpha,
                    min_periods=min_periods,
                    adjust=adjust,
                    ignore_na=ignore_na,
                    axis=axis,
                    times=times,
                    method=method,
                    ).mean()



from datetime import timedelta
from decimal import Decimal

import numpy as np
import pandas as pd
import math

from demeter import DemeterError
from .common import get_real_n


def actual_volatility(data: pd.Series,
                      window: timedelta = timedelta(minutes=5),
                      timeunit: timedelta = timedelta(days=1)) -> "Series(float64)":
    """
    get actual volatility. step:

    1. get window size according to n and unit
    2. get return rate according to window size
    3. calculate standard deviation for each data point, form current point p[n] to previous point p[n - window_size].

    note:
    1. the first window_size * 2 - 1 data will be NAN

    :param data: data to process
    :type data: Series
    :param window: window width
    :type window: timedelta
    :param timeunit: time unit for volatility, default is one day
    :type timeunit: timedelta
    """
    real_n = get_real_n(data, window)

    if real_n * 2 - 1 >= len(data.index):
        raise DemeterError(f"data length is {len(data.index)}, but window size is {real_n}, "
                           f"data length should be greater than {real_n * 2 - 1} to avoid all NAN")

    if data.dtypes == "object":
        shifted = data.shift(periods=real_n, fill_value=Decimal(float("nan")))
        return_rate = data.div(shifted).apply(Decimal.ln)
    else:
        shifted = data.shift(periods=real_n)
        return_rate = data.div(shifted).apply(np.log)

    volatility_column: pd.Series = return_rate.rolling(window=real_n).std()
    amp = math.sqrt(timeunit.total_seconds() / window.total_seconds())
    return volatility_column * amp



import json
from datetime import date

import pandas

from ._typing import ChainType
from .swap_contract import Constant


def download_bigquery_pool_event_oneday(chain: ChainType, contract_address: str, one_date: date) -> "pandas.DataFrame":
    """
    query log data from big_query
    sample response
    [{
          "log_index": "38",
          "transaction_hash": "0xb013afdc4272ccf59a19cfa3943d2af9e818dd3a88981fc0e31e043233d31d1a",
          "transaction_index": "1",
          "address": "0x8ef34625aec17541b6147a40a3840c918263655e",
          "data": "0x0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000002022f034f3ae810000000000000000000000000000000000000000000000c30329a44dfdbf0de60000000000000000000000000000000000000000000000000000000000000000",
          "topics": ["0xd78ad95fa46c994b6551d0da85fc275fe613ce37657fb8d5e3d130840159d822", "0x0000000000000000000000007a250d5630b4cf539739df2c5dacb4c659f2488d", "0x000000000000000000000000700000f7c2c71caab6b250ca85237117ff702ebb"],
          "block_timestamp": "2022-09-01T01:01:52Z",
          "block_number": "15449884",
          "block_hash": "0x48268cc49b9a68fa08a5954a4a841f36205f8ba5a957cb632165a817c6817b05"
    }]
    :param chain:
    :param contract_address:
    :param one_date:
    :return:
    """
    from google.cloud import bigquery
    client = bigquery.Client()
    # no index on blocknum for bigquery. Do not use blocknum in where. slow.
    query = f"""SELECT
        block_timestamp,
        block_number,
        transaction_hash,
        transaction_index,
        log_index,
        topics,
        DATA
        FROM
        {ModuleUtils.get_table_name(chain)}
        WHERE
            (topics[SAFE_OFFSET(0)] = '{Constant.MINT_KECCAK}'
            OR topics[SAFE_OFFSET(0)] = '{Constant.BURN_KECCAK}'
            OR topics[SAFE_OFFSET(0)] = '{Constant.SWAP_KECCAK}' OR topics[SAFE_OFFSET(0)] = '{Constant.COLLECT_KECCAK}')
            AND DATE(block_timestamp) >=  DATE("{one_date}")
            AND DATE(block_timestamp) <=  DATE("{one_date}")
            AND address = "{contract_address}"  order by block_number asc,log_index asc"""
    # print(query);
    query_job = client.query(query)  # Make an API request.
    result: pandas.DataFrame = query_job.to_dataframe(create_bqstorage_client=False)
    return result


class ModuleUtils(object):
    @staticmethod
    def get_table_name(chain_type: ChainType) -> str:
        match chain_type:
            case ChainType.Polygon:
                return "public-data-finance.crypto_polygon.logs"
            case ChainType.Ethereum:
                return "bigquery-public-data.crypto_ethereum.logs"
            case _:
                raise RuntimeError("chain type {} is not supported by BigQuery", chain_type)



from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import NamedTuple

from .utils.application import get_formatted_str

DECIMAL_ZERO = Decimal(0)


class TimeUnitEnum(Enum):
    """
    time unit for moving average,

    * minute
    * hour
    * day
    """
    minute = 1
    hour = 60
    day = 60 * 24


class UnitDecimal(Decimal):
    """
    Decimal with unit, such a 1 eth.

    It's inherit from Decimal, but considering performance issues, calculate function is not override,
    so if you do calculate on this object, return type will be Decimal

    :param number: number to keep
    :type number: Decimal
    :param unit: unit of the number, eg: eth
    :type unit: str
    :param output_format: output format, follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html
    :type output_format: str
    """
    __integral = Decimal(1)
    default_output_format = ".8g"

    def __new__(cls, value, unit: str, output_format=None):
        obj = Decimal.__new__(cls, value)
        obj.unit = unit
        obj.output_format: str = output_format if output_format is not None else UnitDecimal.default_output_format
        return obj

    def to_str(self):
        """
        get formatted string of this decimal. format is defined in self.output_format and unit will be append.

        :return: formatted string
        :rtype: str
        """
        dec = self.quantize(self.__integral) if self == self.to_integral() else self.normalize()
        return "{:{}} {}".format(dec, self.output_format, self.unit)


class TokenInfo(NamedTuple):
    """
    token info

    :param name: token symbol, will be set as unit of a token value
    :type name: str
    :param decimal: decimal of this token
    :type decimal: int
    """
    name: str
    decimal: int


class Asset(NamedTuple):
    """
    asset info of a token

    :param token: token info
    :type token: TokenInfo
    :param amount: amount of this token, eg: 1.123456
    :type amount: Decimal
    """
    token: TokenInfo
    amount: Decimal


class ActionTypeEnum(Enum):
    """
    Trade types

    * add_liquidity = 1,
    * remove_liquidity = 2,
    * buy = 3,
    * sell = 4,
    * collect_fee = 5
    """
    add_liquidity = 1,
    remove_liquidity = 2,
    buy = 3,
    sell = 4,
    collect_fee = 5


class PositionInfo(NamedTuple):
    """
    position information, including tick range and liquidity. It's the immute properties of a position

    :param lower_tick: lower tick
    :type lower_tick: int
    :param upper_tick: upper tick
    :type upper_tick: int

    """
    lower_tick: int
    upper_tick: int

    def __str__(self):
        return f"""tick:{self.lower_tick},{self.upper_tick}"""


BarStatusNames = [
    "base_balance",
    "quote_balance",
    "base_uncollected",
    "quote_uncollected",
    "base_in_position",
    "quote_in_position",
    "net_value",
    "price"
]


@dataclass
class AccountStatus:
    """
    current status of broker

    :param timestamp: timestamp
    :type timestamp: datetime
    :param base_balance: balance of base token
    :type base_balance: UnitDecimal
    :param quote_balance: balance of quote token
    :type quote_balance: UnitDecimal
    :param base_uncollected: base token uncollect fee in all the positions.
    :type base_uncollected: UnitDecimal
    :param quote_uncollected: quote token uncollect fee in all the positions.
    :type quote_uncollected: UnitDecimal
    :param base_in_position: base token amount deposited in positions, calculated according to current price
    :type base_in_position: UnitDecimal
    :param quote_in_position: quote token amount deposited in positions, calculated according to current price
    :type quote_in_position: UnitDecimal
    :param net_value: all the capitals for broker, including balance,uncollected fee, deposited
    :type net_value: UnitDecimal
    :param price: current price
    :type price: UnitDecimal

    """
    timestamp: datetime
    base_balance: UnitDecimal
    quote_balance: UnitDecimal
    base_uncollected: UnitDecimal
    quote_uncollected: UnitDecimal
    base_in_position: UnitDecimal
    quote_in_position: UnitDecimal
    net_value: UnitDecimal
    price: UnitDecimal

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return get_formatted_str({
            "total capital": f"{self.net_value.to_str()}",
            "balance": f"{self.base_balance.to_str()},{self.quote_balance.to_str()}",
            "uncollect fee": f"{self.base_uncollected.to_str()},{self.quote_uncollected.to_str()}",
            "in position amount": f"{self.base_in_position.to_str()},{self.quote_in_position.to_str()}"
        })

    def to_array(self):
        return [
            self.base_balance,
            self.quote_balance,
            self.base_uncollected,
            self.quote_uncollected,
            self.base_in_position,
            self.quote_in_position,
            self.net_value,
            self.price
        ]


class RowData(object):
    """
    data for each bar. Strategy.on_bar() would know which property to use

    :param timestamp: current time of test data
    :param row_id: data index of this line, start with 0, can be used in dataframe.iloc()
    :param netAmount0: net amount of token 0
    :param netAmount1: net amount of token 1
    :param closeTick: last tick of this bar
    :param openTick: first tick of this bar
    :param lowestTick: lowest tick
    :param highestTick: highest tick
    :param inAmount0: swap in amount of token 0
    :param inAmount1: swap in amount of token 1
    :param currentLiquidity: current liquidity
    :param open: first price of this bar
    :param price: latest price of this bar
    :param low: lowest price
    :param high: highest price
    :param volume0: volume of token 0
    :param volume1: volume of token 1
    """

    def __init__(self):
        self.timestamp: datetime = None
        self.row_id: int = None
        self.netAmount0: int = None
        self.netAmount1: int = None
        self.closeTick: int = None
        self.openTick: int = None
        self.lowestTick: int = None
        self.highestTick: int = None
        self.inAmount0: int = None
        self.inAmount1: int = None
        self.currentLiquidity: Decimal = None
        self.open: Decimal = None
        self.price: Decimal = None
        self.low: Decimal = None
        self.high: Decimal = None
        self.volume0: Decimal = None
        self.volume1: Decimal = None


@dataclass
class BaseAction(object):
    """
    Parent class of broker actions,

    :param trade_type: action type
    :type action_type: ActionTypeEnum
    :param timestamp: action time
    :type timestamp: datetime
    :param base_balance_after: after action balance of base token
    :type base_balance_after: UnitDecimal
    :param quote_balance_after: after action balance of quote token
    :type quote_balance_after: UnitDecimal
    """
    action_type: ActionTypeEnum = field(default=False, init=False)
    timestamp: datetime = field(default=False, init=False)
    base_balance_after: UnitDecimal
    quote_balance_after: UnitDecimal

    def get_output_str(self):
        return str(self)


@dataclass
class AddLiquidityAction(BaseAction):
    """
    Add Liquidity

    :param base_amount_max: inputted base token amount, also the max amount to deposit
    :type base_amount_max: ActionTypeEnum
    :param quote_amount_max: inputted base token amount, also the max amount to deposit
    :type quote_amount_max: datetime
    :param lower_quote_price: lower price base on quote token.
    :type lower_quote_price: UnitDecimal
    :param upper_quote_price: upper price base on quote token.
    :type upper_quote_price: UnitDecimal
    :param base_amount_actual: actual used base token
    :type base_amount_actual: UnitDecimal
    :param quote_amount_actual: actual used quote token
    :type quote_amount_actual: UnitDecimal
    :param position: generated position
    :type position: PositionInfo
    :param liquidity: liquidity added
    :type liquidity: int
    """
    base_amount_max: UnitDecimal
    quote_amount_max: UnitDecimal
    lower_quote_price: UnitDecimal
    upper_quote_price: UnitDecimal
    base_amount_actual: UnitDecimal
    quote_amount_actual: UnitDecimal
    position: PositionInfo
    liquidity: int
    action_type = ActionTypeEnum.add_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;31m{"Add liquidity":<20}\033[0m""" + \
               get_formatted_str({
                   "max amount": f"{self.base_amount_max.to_str()},{self.quote_amount_max.to_str()}",
                   "price": f"{self.lower_quote_price.to_str()},{self.upper_quote_price.to_str()}",
                   "position": self.position,
                   "liquidity": self.liquidity,
                   "balance": f"{self.base_balance_after.to_str()}(-{self.base_amount_actual.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_amount_actual.to_str()})"
               })


@dataclass
class CollectFeeAction(BaseAction):
    """
    collect fee

    :param position: position to operate
    :type position: PositionInfo
    :param base_amount: fee collected in base token
    :type base_amount: UnitDecimal
    :param quote_amount: fee collected in quote token
    :type quote_amount: UnitDecimal

    """
    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal
    action_type = ActionTypeEnum.collect_fee

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;33m{"Collect fee":<20}\033[0m""" + \
               get_formatted_str({
                   "position": self.position,
                   "balance": f"{self.base_balance_after.to_str()}(+{self.base_amount.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_amount.to_str()})"
               })


@dataclass
class RemoveLiquidityAction(BaseAction):
    """
    TODO: update
    remove position

    :param position: position to operate
    :type position: PositionInfo
    :param base_amount: base token amount collected
    :type base_amount: UnitDecimal
    :param quote_amount: quote token amount collected
    :type quote_amount: UnitDecimal
    :param removed_liquidity: liquidity number has removed
    :type removed_liquidity: int
    :param remain_liquidity: liquidity number left in position
    :type remain_liquidity: int

    """
    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal
    removed_liquidity: int
    remain_liquidity: int
    action_type = ActionTypeEnum.remove_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;32m{"Remove liquidity":<20}\033[0m""" + \
               get_formatted_str({
                   "position": self.position,
                   "balance": f"{self.base_balance_after.to_str()}(+0), {self.quote_balance_after.to_str()}(+0)",
                   "token_got": f"{self.base_amount.to_str()},{self.quote_amount.to_str()}",
                   "removed liquidity": self.removed_liquidity,
                   "remain liquidity": self.remain_liquidity
               })


@dataclass
class BuyAction(BaseAction):
    """
    buy token, swap from base token to quote token.

    :param amount: amount to buy(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in base token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """
    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal
    action_type = ActionTypeEnum.buy

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;36m{"Buy":<20}\033[0m""" + \
               get_formatted_str({
                   "price": self.price.to_str(),
                   "fee": self.fee.to_str(),
                   "balance": f"{self.base_balance_after.to_str()}(-{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_change.to_str()})"
               })


@dataclass
class SellAction(BaseAction):
    """
    sell token, swap from quote token to base token.

    :param amount: amount to sell(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in quote token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """
    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal
    action_type = ActionTypeEnum.sell

    def get_output_str(self):
        return f"""\033[1;37m{"Sell":<20}\033[0m""" + \
               get_formatted_str({
                   "price": self.price.to_str(),
                   "fee": self.fee.to_str(),
                   "balance": f"{self.base_balance_after.to_str()}(+{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_change.to_str()})"
               })


@dataclass
class EvaluatingIndicator:
    """
    Indicator to evaluate a strategy

    :param annualized_returns: annualized returns
    :type annualized_returns: UnitDecimal
    :param benchmark_returns: benchmark returns
    :type benchmark_returns: UnitDecimal


    """
    annualized_returns: UnitDecimal
    benchmark_returns: UnitDecimal
    max_drawdown: UnitDecimal

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return get_formatted_str({
            "annualized_returns": self.annualized_returns.to_str(),
            "benchmark_returns": self.benchmark_returns.to_str(),
            "max_drawdown": self.max_drawdown.to_str()
        })


class DemeterError(Exception):
    def __init__(self, message):
        self.message = message


class EthError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message



from dataclasses import dataclass
from datetime import datetime
from enum import Enum

DEFAULT_SAVE_PATH = "data"


class DataSource(Enum):
    BigQuery = 1
    RPC = 2


class ChainType(Enum):
    Ethereum = 1
    Polygon = 2
    Optimism = 3
    Arbitrum = 4
    Celo = 5


@dataclass
class RpcParam:
    end_point: str
    start_height: int
    end_height: int
    auth_string: str = ""
    proxy: str = ""
    batch_size: int = 500


@dataclass
class BigQueryParam:
    auth_file: str
    start: str
    end: str = datetime.now().strftime("%Y-%m-%d")


@dataclass
class DownloadParam:
    chain: ChainType = ChainType.Ethereum
    source: DataSource = DataSource.BigQuery
    pool_address: str = ""
    save_path: str = DEFAULT_SAVE_PATH
    rpc = RpcParam("", 0, 0)
    big_query = BigQueryParam("", "")


class OnchainTxType(Enum):
    MINT = 0
    SWAP = 2
    BURN = 1
    COLLECT = 3


MarketDataNames = [
    "timestamp",
    "netAmount0",
    "netAmount1",
    "closeTick",
    "openTick",
    "lowestTick",
    "highestTick",
    "inAmount0",
    "inAmount1",
    "currentLiquidity",
]


class MarketData(object):

    def __init__(self):
        self.timestamp = None
        self.netAmount0 = 0
        self.netAmount1 = 0
        self.closeTick = None
        self.openTick = None
        self.lowestTick = None
        self.highestTick = None
        self.inAmount0 = 0
        self.inAmount1 = 0
        self.currentLiquidity = None

    def to_array(self):
        return [
            self.timestamp,
            self.netAmount0,
            self.netAmount1,
            self.closeTick,
            self.openTick,
            self.lowestTick,
            self.highestTick,
            self.inAmount0,
            self.inAmount1,
            self.currentLiquidity
        ]

    def __str__(self):
        return str(self.timestamp)

    def fill_missing_field(self, prev_data) -> bool:
        """
        fill missing field with previous data
        :param prev_data:
        :return: data is available or not
        """
        if prev_data is None:
            prev_data = MarketData()
        self.closeTick = self.closeTick if self.closeTick is not None else prev_data.closeTick
        self.openTick = self.openTick if self.openTick is not None else prev_data.closeTick
        self.lowestTick = self.lowestTick if self.lowestTick is not None else prev_data.closeTick
        self.highestTick = self.highestTick if self.highestTick is not None else prev_data.closeTick
        self.currentLiquidity = self.currentLiquidity if self.currentLiquidity is not None else prev_data.currentLiquidity

        return False if (self.closeTick is None or self.currentLiquidity is None) else True



from enum import Enum

import pandas as pd
from pandas import _typing as pd_typing

from ._typing import DemeterError

DEFAULT_AGG_METHOD = "first"


class LineTypeEnum(Enum):
    """
    predefined column, used to define fillna method.
    """
    timestamp = 1
    netAmount0 = 2
    netAmount1 = 3
    closeTick = 4
    openTick = 5
    lowestTick = 6
    highestTick = 7
    inAmount0 = 8
    inAmount1 = 9
    currentLiquidity = 10
    other = 100

    @staticmethod
    def safe_get(key):
        try:
            return LineTypeEnum[key]
        except KeyError:
            return None


def get_agg_by_type(line_type: LineTypeEnum) -> str:
    match line_type:
        case LineTypeEnum.openTick:
            return "first"
        case LineTypeEnum.closeTick:
            return "last"
        case LineTypeEnum.highestTick:
            return "max"
        case LineTypeEnum.lowestTick:
            return "min"
        case LineTypeEnum.netAmount0 | LineTypeEnum.netAmount1 | LineTypeEnum.inAmount0 | LineTypeEnum.inAmount1 | LineTypeEnum.currentLiquidity:
            return "sum"
        case _:
            return ""


def get_fillna_param(line_type: LineTypeEnum, method=None, value=0):
    match line_type:
        case LineTypeEnum.openTick | LineTypeEnum.closeTick | LineTypeEnum.highestTick | LineTypeEnum.lowestTick | LineTypeEnum.currentLiquidity:
            method = "ffill"
            value = None
        case LineTypeEnum.netAmount0 | LineTypeEnum.netAmount1 | LineTypeEnum.inAmount0 | LineTypeEnum.inAmount1:
            method = None
            value = 0
    return {"method": method, "value": value}


class Cursorable(object):
    """
    give Lines or line a cursor. which help to access current row with index 0 in a loop.  eg:

    if data= [1,2,3,4,5], self.cursor=0

    * data.get_by_cursor[0]==1
    * data.get_by_cursor[2]==3
    * data.get_by_cursor[4]==5

    if self.cursor=2

    * data.get_by_cursor[-2]==1, access previse row with negative index,
    * data.get_by_cursor[0]==3, access current row with 0,
    * data.get_by_cursor[2]==5, access after row with positive index

    if self.cursor=4

    * data.get_by_cursor[-4]==1
    * data.get_by_cursor[-2]==3
    * data.get_by_cursor[0]==5
    """

    def __init__(self):
        self.cursor = 0

    def row_length(self):
        """
        get total length

        """
        return self.index.size

    def get_cursor_range(self):
        """
        get legal range for index
        :return: min, max
        """
        return -self.cursor, self.index.size - self.cursor - 1

    def move_cursor_to_next(self):
        """
        move cursor to on_bar index
        """
        self.cursor += 1

    def reset_cursor(self):
        """
        reset cursor to zero
        """
        self.cursor = 0

    def get_by_cursor(self, i):
        """
         access row by cursor

         :param i: index
         :type i: int
         :return: selected row, in form of pd.Series
         """
        real_index = self.cursor + i

        if real_index < 0 or real_index >= self.index.size:
            raise IndexError("index out of range")
        return self.iloc[real_index]


class Line(pd.Series, Cursorable):
    """
    a column of data, inherit from pandas.Series

    and it has two additional function:

    * it has a property of column type. if the column type is predefined, then fillna and resample method is not changeable
    * it provide cursor.

    note: index of this Series must be datetime
    """

    def __init__(
            self,
            data=None,
            index=None,
            dtype: pd_typing.Dtype | None = None,
            name=None,
            copy: bool = False,
            fastpath: bool = False,
            line_type=LineTypeEnum.other
    ):

        if not isinstance(index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("index must be datetime")
        super().__init__(data, index, dtype, name, copy, fastpath)
        self.line_type = line_type
        Cursorable.__init__(self)

    def resample_by_type(
            self,
            rule,
            axis=0,
            closed: str | None = None,
            label: str | None = None,
            convention: str = "start",
            kind: str | None = None,
            loffset=None,
            base: int | None = None,
            on=None,
            level=None,
            origin: str | pd_typing.TimestampConvertibleTypes = "start_day",
            offset: pd_typing.TimedeltaConvertibleTypes | None = None,
            agg=""
    ) -> pd.Series:
        """
        with this method, predefined type will be resampled correctly, other param is the same to Series.resample()

        :param agg: if line_type is other, should define resample method, such as sum, mean
        :return: resampled data
        """
        resampler = super().resample(rule, axis, closed, label, convention, kind, loffset, base, on, level, origin,
                                     offset)
        agg_method = get_agg_by_type(self.line_type)
        if agg_method == "":
            agg_method = DEFAULT_AGG_METHOD
        if self.line_type == LineTypeEnum.other:
            if agg == "":
                agg_method = DEFAULT_AGG_METHOD
            else:
                agg_method = agg
        return resampler.agg(agg_method)

    def fillna(
            self,
            value: object | pd_typing.ArrayLike | None = None,
            method: pd_typing.FillnaOptions | None = None,
            axis=None,
            inplace=False,
            limit=None,
            downcast=None,
    ) -> pd.Series:
        """
        fill empty item. param is the same to pandas.Series.fillna

        if line type is predefined, method and value will be omitted, and data will be filled as predefined

        """
        param = get_fillna_param(self.line_type, method, value)
        method = param["method"]
        value = param["value"]
        return super().fillna(
            value=value,
            method=method,
            axis=axis,
            inplace=inplace,
            limit=limit,
            downcast=downcast,
        )

    @staticmethod
    def from_series(series: pd.Series, row_index: 0) -> "Line":
        """
        convert serise to line, line type will be defined by series.name

        :param series: series to process
        :type series: pd.Series
        :param row_index: new row index
        :type row_index: int
        :return: converted Line
        :rtype: Line
        """
        line_type = LineTypeEnum.other
        try:
            line_type = LineTypeEnum[series.name]
        except:
            line_type = LineTypeEnum.other
        line = series.copy(deep=True)
        line.cursor = row_index
        setattr(line, "line_type", line_type)
        return line


class Lines(pd.DataFrame, Cursorable):
    """
    a data table, inherit from pandas.DataFrame

    and it has two additional function:

    * it has a property of column type. if the column type is predefined, then fillna and resample method can not change
    * it provide cursor. assistant to access current row in a loop

    note: index of this DataFrame must be datetime
    """

    def __init__(
            self,
            data=None,
            index: pd_typing.Axes | None = None,
            columns: pd_typing.Axes | None = None,
            dtype: pd_typing.Dtype | None = None,
            copy: bool | None = None,
    ):
        if (index is not None) and (not isinstance(index, pd.core.indexes.datetimes.DatetimeIndex)):
            raise DemeterError("index must be datetime")
        # add line by type
        line_data = dict()
        if data is not None:
            if isinstance(data, list):
                for item in data:
                    Lines.__set_line_name_by_object(item, line_data)
            # after line is added, it's type will become series, so we identify its type by name
            elif isinstance(data, dict):
                line_data = data
            else:
                Lines.__set_line_name_by_object(data, line_data)
        super().__init__(line_data, index, columns, dtype, copy)
        Cursorable.__init__(self)

    @staticmethod
    def __set_line_name_by_object(item, line_data):
        """
        after line is added, it's type will become series, so we set their type to name.
        """
        if isinstance(item, Line) and item.line_type != LineTypeEnum.other:
            line_data[item.line_type.name] = item
        elif isinstance(item, pd.Series) and item.name:
            line_data[item.name] = item
        else:
            line_data["Column" + str(len(line_data.keys()))] = item

    def append_column(self, name: str, column: Line):
        """
        add a column to Lines

        :param name: column name
        :type name: str
        :param column: column to add, type must be line, or pandas.Series, but index of Series should be datetime
        :type: Line

        """
        if not isinstance(column.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("index must be datetime")
        if self.data.columns.size == 0:
            self.data.index = column.index
        self.data[name] = column

    def fillna(
            self,
            value: object | pd_typing.ArrayLike | None = None,
            method: pd_typing.FillnaOptions | None = None,
            axis: pd_typing.Axis | None = None,
            inplace: bool = False,
            limit=None,
            downcast=None,
    ) -> pd.DataFrame | None:
        """
        fill empty item. param is the same to pandas.Series.fillna

        if column name is predefined, method and value will be omitted, and data will be filled as predefined

        """
        new_df = self.copy(False)

        # fill close tick first, it will be used later.
        if LineTypeEnum.closeTick.name in new_df.columns:
            new_df[LineTypeEnum.closeTick.name] = new_df[LineTypeEnum.closeTick.name].fillna(value=None, method="ffill",
                                                                                             axis=axis,
                                                                                             inplace=inplace,
                                                                                             limit=limit,
                                                                                             downcast=downcast)
        for column_name in new_df.columns:
            if column_name == LineTypeEnum.closeTick.name:
                continue
            line_type = LineTypeEnum.safe_get(column_name)
            if line_type == "":
                new_df[column_name] = new_df[column_name].fillna(value, method, axis, inplace, limit, downcast)
            else:
                param = get_fillna_param(line_type, method, value)
                current_method = param["method"]
                current_value = param["value"]
                # all tick related field will be filled with close_tick.
                if line_type in [LineTypeEnum.openTick, LineTypeEnum.highestTick,
                                 LineTypeEnum.lowestTick] and LineTypeEnum.closeTick.name in new_df.columns:
                    current_method = None
                    current_value = new_df[LineTypeEnum.closeTick.name]
                new_df[column_name] = new_df[column_name].fillna(value=current_value, method=current_method, axis=axis,
                                                                 inplace=inplace,
                                                                 limit=limit,
                                                                 downcast=downcast)
        new_lines = Lines.from_dataframe(new_df)
        new_lines.cursor = self.cursor
        return new_lines

    def resample_by_type(
            self,
            rule,
            axis=0,
            closed: str | None = None,
            label: str | None = None,
            convention: str = "start",
            kind: str | None = None,
            loffset=None,
            base: int | None = None,
            on=None,
            level=None,
            origin: str | pd_typing.TimestampConvertibleTypes = "start_day",
            offset: pd_typing.TimedeltaConvertibleTypes | None = None,
            agg=dict()
    ) -> "Lines":
        """
        with this method, predefined column will be resampled correctly, other param is the same to Series.resample()

        :param agg: if line name is not predefined, should define resample method, such as sum, mean
        :return: resampled data
        """
        resampler = super().resample(rule, axis, closed, label, convention, kind, loffset, base, on, level, origin,
                                     offset)
        agg_dict = dict()
        for column_name in self.columns:
            column_type = LineTypeEnum.safe_get(column_name)
            agg_method = get_agg_by_type(column_type)
            if agg_method == "":
                if column_name in agg:
                    agg_method = agg[column_name]
                else:
                    agg_method = DEFAULT_AGG_METHOD
            agg_dict[column_name] = agg_method
        df = resampler.agg(agg_dict)
        new_lines = Lines.from_dataframe(df)
        new_lines.cursor = 0
        return new_lines

    def get_line(self, index: int = None, name: str = None) -> Line:
        """
        get a column of Lines

        :param index: column index, if index is set, name param will be override
        :type index: int
        :param name: column name
        :type name: str
        :return: selected column
        :rtype: Line
        """
        if not index and not name:
            raise DemeterError("must use index or name")
        if index:
            return Line.from_series(self.iloc[:, index], self.cursor)
        elif name:
            return Line.from_series(self[name], self.cursor)

    @staticmethod
    def from_dataframe(df: pd.DataFrame) -> "Lines":
        """
        convert pandas.DataFrame to Lines. this method will copy inner variable to new Lines.

        :param df: dataframe to convert
        :type df: DataFrame
        :return: Lines
        :rtype: Lines
        """
        new_lines = Lines(index=df.index)
        new_lines._mgr = df._mgr
        return new_lines

    @staticmethod
    def load_downloaded(path: str) -> "Lines":
        """
        load data which saved by downloader as lines
        :param path: csv file path
        :type path: str
        :return: loaded data
        :rtype: Lines
        """
        df = pd.read_csv(path)
        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)
        return Lines.from_dataframe(df)



import csv
import datetime
import operator
import os.path
import pickle
import time
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from operator import itemgetter
from .eth_req import EthRequestClient, GetLogsParam
from tqdm import tqdm  # process bar
import json
from ._typing import DownloadParam, ChainType
from .swap_contract import Constant
from .utils import get_file_name

height_cache_file_name = "_height_timestamp.pkl"
# preserved for future extension. such as special logic for each chain.
CHAINS = {
    ChainType.Polygon: {
        "is_poa": True
    },
    ChainType.Arbitrum: {
        "is_poa": True
    },
    ChainType.Optimism: {
        "is_poa": True
    },
    ChainType.Ethereum: {
        "is_poa": False
    },
}


@dataclass
class ContractConfig:
    address: str
    topics: list[str]
    batch_size: int
    one_by_one: bool


def save_block_dict(height_cache_path, block_dict):
    with open(height_cache_path, 'wb') as f:
        pickle.dump(block_dict, f)


# def dict_factory(cursor, row):
#     d = {}
#     for idx, col in enumerate(cursor.description):
#         d[col[0]] = row[idx]
#     return d


def cut(obj, sec):
    return [obj[i:i + sec] for i in range(0, len(obj), sec)]


def fill_block_info(log, client: EthRequestClient, block_dict):
    height = log['block_number']
    if height not in block_dict:
        block_dt = client.get_block_timestamp(height)
        block_dict[height] = block_dt
    log['block_timestamp'] = block_dict[height].isoformat()
    log['block_dt'] = block_dict[height]
    return log


def query_event_log(client: EthRequestClient, contract_config: ContractConfig, start_height: int, end_height: int,
                    save_path: str,
                    block_dict, chain):
    collect_dt, log_by_day_dict, collect_start = None, OrderedDict(), None  # collect date, date log by day，collect start time
    start_tp = time.time()
    downloaded_day = []
    with tqdm(total=(end_height - start_height + 1), ncols=150) as pbar:
        for height_slice in cut([i for i in range(start_height, end_height + 1)], contract_config.batch_size):
            start = height_slice[0]
            end = height_slice[-1]
            if contract_config.one_by_one:
                logs = []
                for topic_hex in contract_config.topics:
                    tmp_logs = client.get_logs(GetLogsParam(contract_config.address,
                                                            start,
                                                            end,
                                                            [topic_hex]))
                    logs.extend(tmp_logs)
            else:
                logs = client.get_logs(GetLogsParam(contract_config.address, start, end, None))
            log_lst = []
            for log in logs:
                log['blockNumber'] = int(log['blockNumber'], 16)
                if len(log['topics']) > 0 and (log['topics'][0] in contract_config.topics):
                    if log["removed"]:
                        continue
                    log_lst.append({
                        'block_number': log['blockNumber'],
                        'transaction_hash': log['transactionHash'],
                        'transaction_index': log['transactionIndex'],
                        'log_index': log['logIndex'],
                        'DATA': log["data"],
                        'topics': json.dumps(log['topics'])
                    })
            with ThreadPoolExecutor(max_workers=10) as t:
                obj_lst = []
                for data in log_lst:
                    obj = t.submit(fill_block_info, data, client, block_dict)
                    obj_lst.append(obj)
                for future in as_completed(obj_lst):
                    data = future.result()
                    block_time_by_day = data['block_dt'].date()
                    data.pop('block_dt')
                    if block_time_by_day not in log_by_day_dict:
                        log_by_day_dict[block_time_by_day] = [data]
                    else:
                        log_by_day_dict[block_time_by_day].append(data)

            pbar.update(n=len(height_slice))
            if (len(height_slice) < contract_config.batch_size) or (
                    len(height_slice) >= contract_config.batch_size and len(log_by_day_dict) >= 2):
                # collect_dt/collect_start
                if len(log_by_day_dict) > 0:
                    log_by_day_dict = OrderedDict(sorted(log_by_day_dict.items(), key=operator.itemgetter(0)))

                    collect_dt, one_day_data = log_by_day_dict.popitem(last=False)
                    one_day_data = sorted(one_day_data,
                                          key=itemgetter('block_number', 'transaction_index', 'log_index'))
                    collect_dt = collect_dt.strftime('%Y-%m-%d')
                    collect_start = one_day_data[0]['block_number']
                    # write to csv
                    save_one_day(save_path, collect_dt, contract_config, one_day_data, chain)
                    downloaded_day.append(collect_dt)
                    print(f'\nsaved date: {collect_dt}, start height: {collect_start}, '
                          f'length: {len(one_day_data)}, time: {time.time() - start_tp} s')
                    start_tp = time.time()
        return downloaded_day


def save_one_day(save_path, collect_dt, contract_config, one_day_data, chain: ChainType):
    with open(get_file_name(save_path, chain.name, contract_config.address, collect_dt, True), 'w') as csvfile:
        writer = csv.DictWriter(csvfile,
                                fieldnames=['block_number', 'block_timestamp', 'transaction_hash',
                                            'transaction_index',
                                            'log_index', 'DATA', 'topics'])
        writer.writeheader()
        for item in one_day_data:
            writer.writerow(item)


def download_and_save_by_day(config: DownloadParam):
    chain_config = CHAINS[config.chain]
    height_cache_path = config.chain.name + height_cache_file_name
    if os.path.exists(height_cache_path):
        with open(height_cache_path, 'rb') as f:
            block_dict = pickle.load(f)
            print(f"Height cache has loaded, length: {len(block_dict)}")
    else:
        block_dict: dict[int:datetime.datetime] = {}

    client = EthRequestClient(config.rpc.end_point, config.rpc.proxy, config.rpc.auth_string)

    try:
        downloaded_day = query_event_log(client,
                                         ContractConfig(config.pool_address,
                                                        [Constant.SWAP_KECCAK,
                                                         Constant.BURN_KECCAK,
                                                         Constant.COLLECT_KECCAK,
                                                         Constant.MINT_KECCAK],
                                                        config.rpc.batch_size,
                                                        False),
                                         config.rpc.start_height,
                                         config.rpc.end_height,
                                         config.save_path,
                                         block_dict,
                                         config.chain)
    except Exception as e:
        print(e)
        import traceback
        print(traceback.format_exc())
    print(f"saving height cache, length {len(block_dict)}")
    save_block_dict(height_cache_path, block_dict)
    return downloaded_day if downloaded_day else []



import math
from decimal import Decimal
from .liquitidymath import get_sqrt_ratio_at_tick


def _x96_to_decimal(number: int):
    return Decimal(number) / 2 ** 96


def decimal_to_x96(number: Decimal):
    return int(Decimal(number) * 2 ** 96)


def _x96_sqrt_to_decimal(sqrt_priceX96, token_decimal_diff=12):
    price = _x96_to_decimal(sqrt_priceX96)
    return (price ** 2) / 10 ** token_decimal_diff


## can round by spacing?
def sqrt_price_to_tick(sqrt_priceX96: int) -> int:
    decimal_price = _x96_to_decimal(sqrt_priceX96)
    return pool_price_to_tick(decimal_price)


def pool_price_to_tick(price_decimal: Decimal):
    return int(math.log(price_decimal, math.sqrt(1.0001)))


def tick_to_sqrtPriceX96(tick: int):
    return get_sqrt_ratio_at_tick(tick)


def tick_to_quote_price(tick: int, token_0_decimal, token_1_decimal, is_token0_base: bool):
    sqrt_price = get_sqrt_ratio_at_tick(tick)
    decimal_price = _x96_to_decimal(sqrt_price) ** 2
    pool_price = decimal_price * Decimal(10 ** (token_0_decimal - token_1_decimal))
    return Decimal(1 / pool_price) if is_token0_base else pool_price


def quote_price_to_tick(based_price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token_base) -> int:
    # quote price->add decimal pool price->sqrt_price ->ticker
    sqrt_price = quote_price_to_sqrt(based_price, token_0_decimal, token_1_decimal, is_token_base)
    tick = sqrt_price_to_tick(sqrt_price)
    return tick


def quote_price_to_sqrt(based_price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token_base) -> int:
    # quote price->add decimal pool price->sqrt_price ->ticker

    price = 1 / based_price if is_token_base else based_price
    pool_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    decimal_price = Decimal.sqrt(pool_price)
    return decimal_to_x96(decimal_price)


def from_wei(token_amt: int, decimal: int):
    return Decimal(token_amt) / Decimal(10 ** decimal)



from datetime import datetime
from decimal import Decimal
from typing import Union

from .helper import tick_to_quote_price, quote_price_to_tick, quote_price_to_sqrt, tick_to_sqrtPriceX96
from .liquitidymath import get_sqrt_ratio_at_tick
from .types import PoolBaseInfo, TokenInfo, BrokerAsset, Position, PoolStatus
from .v3_core import V3CoreLib
from .._typing import PositionInfo, DemeterError, AddLiquidityAction, RemoveLiquidityAction, BuyAction, SellAction, \
    CollectFeeAction, AccountStatus, DECIMAL_ZERO, UnitDecimal
from ..utils.application import float_param_formatter


class Broker(object):
    """
    Broker manage assets in back testing. Including asset, positions. it also provides operations for positions,
    such as add/remove liquidity, swap assets.

    :param pool_info: pool information
    :type pool_info: PoolBaseInfo
    """

    def __init__(self, pool_info: PoolBaseInfo):
        self._pool_info = pool_info
        # init balance
        self._is_token0_base = pool_info.is_token0_base
        self._asset0 = BrokerAsset(pool_info.token0, DECIMAL_ZERO)
        self._asset1 = BrokerAsset(pool_info.token1, DECIMAL_ZERO)
        base_asset, quote_asset = self.__convert_pair(self._asset0, self._asset1)
        self._base_asset: BrokerAsset = base_asset
        self._quote_asset: BrokerAsset = quote_asset
        self._init_amount0 = DECIMAL_ZERO
        self._init_amount1 = DECIMAL_ZERO
        # status
        self._positions: dict[PositionInfo:Position] = {}
        self._pool_status = PoolStatus(None, 0, DECIMAL_ZERO, DECIMAL_ZERO, DECIMAL_ZERO, DECIMAL_ZERO)
        self._price_unit = f"{self.base_asset.name}/{self.quote_asset.name}"
        # internal temporary variable
        self.action_buffer = []

    def __str__(self):
        return f"Pool: {self.pool_info}, position count: {len(self.positions)}, " \
               f"balance: {self.base_asset.balance}{self.base_asset.name},{self.quote_asset.balance}{self.quote_asset.name}"

    @property
    def positions(self) -> dict[PositionInfo:Position]:
        """
        current positions in broker

        :return: all positions
        :rtype: dict[PositionInfo:Position]
        """
        return self._positions

    @property
    def pool_info(self) -> PoolBaseInfo:
        """
        Get pool info.

        :return: pool info
        :rtype: PoolBaseInfo
        """
        return self._pool_info

    @property
    def asset0(self) -> BrokerAsset:
        """
        get asset 0 info, including balance

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._asset0

    @property
    def asset1(self) -> BrokerAsset:
        """
        get asset 1 info, including balance

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._asset1

    @property
    def base_asset(self) -> BrokerAsset:
        """
        base asset, defined by pool info. It's the reference of asset0 or asset1

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._base_asset

    @property
    def quote_asset(self) -> BrokerAsset:
        """
        quote asset, defined by pool info. It's the reference of asset0 or asset1

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._quote_asset

    @property
    def pool_status(self) -> PoolStatus:
        """
        current pool status. will be writen by runner.

        :return: PoolStatus
        :rtype: PoolStatus
        """
        return self._pool_status

    @pool_status.setter
    def pool_status(self, value: PoolStatus):
        """
        current pool status. will be writen by runner.

        Do not update this property unless necessary

        :param value: current pool status
        :type value: PoolStatus
        """
        self._pool_status = value

    def position(self, position_info: PositionInfo) -> Position:
        """
        get position by position information

        :param position_info: position information
        :type position_info: PositionInfo
        :return: Position
        :rtype: Position
        """
        return self._positions[position_info]

    def __convert_pair(self, any0, any1):
        """
        convert order of token0/token1 to base_token/quote_token, according to self.is_token0_base.

        Or convert order of base_token/quote_token to token0/token1

        :param any0: token0 or any property of token0, eg. balance...
        :param any1: token1 or any property of token1, eg. balance...
        :return: (base,qoute) or (token0,token1)
        """
        return (any0, any1) if self._is_token0_base else (any1, any0)

    @float_param_formatter
    def set_asset(self, token: TokenInfo, amount: Union[Decimal, float]):
        """
        set initial balance for token

        :param token: which token to set
        :type token: TokenInfo
        :param amount: balance, eg: 1.2345
        :type amount: Union[Decimal, float]
        """
        if not self._pool_info:
            raise DemeterError("set up pool info first")
        if token == self._asset0.token_info:
            self._asset0.balance = amount
            self._init_amount0 = amount
        elif token == self._asset1.token_info:
            self._asset1.balance = amount
            self._init_amount1 = amount
        else:
            raise DemeterError("unknown token")

    def update(self):
        """
        re-calculate status
        """
        self.__update_fee()

    def __update_fee(self):
        """
        update fee in all positions according to current status

        fee will be calculated by liquidity
        """
        for position_info, position in self._positions.items():
            V3CoreLib.update_fee(self.pool_info, position_info, position, self.pool_status)

    def get_init_account_status(self, init_price: Decimal, timestamp: datetime = None) -> AccountStatus:
        """
        Get initial status, which will be saved before running any test.

        :param timestamp: timestamp
        :type timestamp: datetime
        :param init_price: beginning price of testing, usually the price in the first item of data array
        :type init_price: Decimal
        :return: status
        :rtype: AccountStatus

        """
        base_init_amount, quote_init_amount = self.__convert_pair(self._init_amount0, self._init_amount1)
        capital = base_init_amount + quote_init_amount * init_price
        return AccountStatus(timestamp,
                             UnitDecimal(base_init_amount, self.base_asset.name),
                             UnitDecimal(quote_init_amount, self.quote_asset.name),
                             UnitDecimal(DECIMAL_ZERO, self.base_asset.name),
                             UnitDecimal(DECIMAL_ZERO, self.quote_asset.name),
                             UnitDecimal(DECIMAL_ZERO, self.base_asset.name),
                             UnitDecimal(DECIMAL_ZERO, self.quote_asset.name),
                             UnitDecimal(capital, self.base_asset.name),
                             UnitDecimal(init_price, self._price_unit))

    def get_account_status(self, price: Decimal = None, timestamp: datetime = None) -> AccountStatus:
        """
        get current status, including positions, balances

        :param price: current price, used for calculate position value and net value, if set to None, will use price in current status
        :type price: Decimal
        :param timestamp: current timestamp, default is none
        :type timestamp: datetime
        :return: BrokerStatus
        """
        if price is None:
            price = self.pool_status.price
        base_asset, quote_asset = self.__convert_pair(self._asset0, self._asset1)
        base_fee_sum = DECIMAL_ZERO
        quote_fee_sum = DECIMAL_ZERO
        sqrt_price = quote_price_to_sqrt(price, self.asset0.decimal, self.asset1.decimal, self._is_token0_base)
        deposit_amount0 = deposit_amount1 = Decimal(0)
        for position_info, position in self._positions.items():
            base_fee, quote_fee = self.__convert_pair(position.pending_amount0,
                                                      position.pending_amount1)
            base_fee_sum += base_fee
            quote_fee_sum += quote_fee
            amount0, amount1 = V3CoreLib.get_token_amounts(self._pool_info, position_info, sqrt_price,
                                                           position.liquidity)
            deposit_amount0 += amount0
            deposit_amount1 += amount1

        base_deposit_amount, quote_deposit_amount = self.__convert_pair(deposit_amount0, deposit_amount1)
        net_value = (base_asset.balance + base_fee_sum + base_deposit_amount) + \
                    (quote_asset.balance + quote_fee_sum + quote_deposit_amount) * price

        return AccountStatus(timestamp=timestamp,
                             base_balance=UnitDecimal(base_asset.balance, self.base_asset.name),
                             quote_balance=UnitDecimal(quote_asset.balance, self.quote_asset.name),
                             base_uncollected=UnitDecimal(base_fee_sum, self.base_asset.name),
                             quote_uncollected=UnitDecimal(quote_fee_sum, self.quote_asset.name),
                             base_in_position=UnitDecimal(base_deposit_amount, self.base_asset.name),
                             quote_in_position=UnitDecimal(quote_deposit_amount, self.quote_asset.name),
                             net_value=UnitDecimal(net_value, self.base_asset.name),
                             price=UnitDecimal(price, self._price_unit))

    def tick_to_price(self, tick: int) -> Decimal:
        """
        convert tick to price

        :param tick: tick
        :type tick: int
        :return: price
        :rtype: Decimal
        """
        return tick_to_quote_price(int(tick), self.pool_info.token0.decimal, self.pool_info.token1.decimal,
                                   self._is_token0_base)

    @float_param_formatter
    def price_to_tick(self, price: Union[Decimal, float]) -> int:
        """
        convert price to tick

        :param price: price
        :type price:  Union[Decimal, float]
        :return: tick
        :rtype: int
        """
        return quote_price_to_tick(price, self.pool_info.token0.decimal, self.pool_info.token1.decimal,
                                   self._is_token0_base)

    def _add_liquidity_by_tick(self, token0_amount: Decimal,
                               token1_amount: Decimal,
                               lower_tick: int,
                               upper_tick: int,
                               sqrt_price_x96: int = -1):
        lower_tick = int(lower_tick)
        upper_tick = int(upper_tick)
        sqrt_price_x96 = int(sqrt_price_x96)

        if sqrt_price_x96 == -1:
            # self.current_tick must be initialed
            sqrt_price_x96 = get_sqrt_ratio_at_tick(self.pool_status.current_tick)
        if lower_tick > upper_tick:
            raise DemeterError("lower tick should be less than upper tick")

        token0_used, token1_used, liquidity, position_info = V3CoreLib.new_position(self._pool_info,
                                                                                    token0_amount,
                                                                                    token1_amount,
                                                                                    lower_tick,
                                                                                    upper_tick,
                                                                                    sqrt_price_x96)
        if position_info in self._positions:
            self._positions[position_info].liquidity += liquidity
        else:
            self._positions[position_info] = Position(DECIMAL_ZERO, DECIMAL_ZERO, liquidity)
        self._asset0.sub(token0_used)
        self._asset1.sub(token1_used)
        return position_info, token0_used, token1_used, liquidity

    def __remove_liquidity(self, position: PositionInfo, liquidity: int = None, sqrt_price_x96: int = -1):
        sqrt_price_x96 = int(sqrt_price_x96) if sqrt_price_x96 != -1 else get_sqrt_ratio_at_tick(
            self.pool_status.current_tick)
        delta_liquidity = liquidity if (liquidity is not None) and liquidity < self.positions[position].liquidity \
            else self.positions[position].liquidity
        token0_get, token1_get = V3CoreLib.close_position(self._pool_info, position, delta_liquidity, sqrt_price_x96)

        self._positions[position].liquidity = self.positions[position].liquidity - delta_liquidity
        self._positions[position].pending_amount0 += token0_get
        self._positions[position].pending_amount1 += token1_get

        return token0_get, token1_get, delta_liquidity

    def __collect_fee(self, position: Position, max_collect_amount0: Decimal = None,
                      max_collect_amount1: Decimal = None):
        token0_fee = max_collect_amount0 if max_collect_amount0 is not None and max_collect_amount0 < position.pending_amount0 else position.pending_amount0
        token1_fee = max_collect_amount1 if max_collect_amount1 is not None and max_collect_amount1 < position.pending_amount1 else position.pending_amount1

        position.pending_amount0 -= token0_fee
        position.pending_amount1 -= token1_fee
        # add un_collect fee to current balance
        self._asset0.add(token0_fee)
        self._asset1.add(token1_fee)

        return token0_fee, token1_fee

    # action for strategy

    @float_param_formatter
    def add_liquidity(self,
                      lower_quote_price: Union[Decimal, float],
                      upper_quote_price: Union[Decimal, float],
                      base_max_amount: Union[Decimal, float] = None,
                      quote_max_amount: Union[Decimal, float] = None,
                      ) -> (PositionInfo, Decimal, Decimal):
        """

        add liquidity, then get a new position

        :param lower_quote_price: lower price base on quote token.
        :type lower_quote_price: Union[Decimal, float]
        :param upper_quote_price: upper price base on quote token.
        :type upper_quote_price: Union[Decimal, float]
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Union[Decimal, float]
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Union[Decimal, float]
        :return: added position, base token used, quote token used
        :rtype: (PositionInfo, Decimal, Decimal)
        """
        base_max_amount = self.base_asset.balance if base_max_amount is None else base_max_amount
        quote_max_amount = self.quote_asset.balance if quote_max_amount is None else quote_max_amount

        token0_amt, token1_amt = self.__convert_pair(base_max_amount, quote_max_amount)
        lower_tick, upper_tick = V3CoreLib.quote_price_pair_to_tick(self._pool_info, lower_quote_price,
                                                                    upper_quote_price)
        lower_tick, upper_tick = self.__convert_pair(upper_tick, lower_tick)
        (created_position, token0_used, token1_used, liquidity) = self._add_liquidity_by_tick(token0_amt,
                                                                                              token1_amt,
                                                                                              lower_tick,
                                                                                              upper_tick)
        base_used, quote_used = self.__convert_pair(token0_used, token1_used)
        self.action_buffer.append(
            AddLiquidityAction(base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                               quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                               base_amount_max=UnitDecimal(base_max_amount, self.base_asset.name),
                               quote_amount_max=UnitDecimal(quote_max_amount, self.quote_asset.name),
                               lower_quote_price=UnitDecimal(lower_quote_price, self._price_unit),
                               upper_quote_price=UnitDecimal(upper_quote_price, self._price_unit),
                               base_amount_actual=UnitDecimal(base_used, self.base_asset.name),
                               quote_amount_actual=UnitDecimal(quote_used, self.quote_asset.name),
                               position=created_position,
                               liquidity=int(liquidity)))
        return created_position, base_used, quote_used, liquidity

    def add_liquidity_by_tick(self, lower_tick: int,
                              upper_tick: int,
                              base_max_amount: Union[Decimal, float] = None,
                              quote_max_amount: Union[Decimal, float] = None,
                              sqrt_price_x96: int = -1,
                              tick: int = -1):
        """

        add liquidity, you need to set tick instead of price.

        :param lower_tick: lower tick
        :type lower_tick: int
        :param upper_tick: upper tick
        :type upper_tick: int
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Union[Decimal, float]
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Union[Decimal, float]
        :param tick: tick price.  if set to none, it will be calculated from current price.
        :type tick: int
        :param sqrt_price_x96: precise price.  if set to none, it will be calculated from current price. this param will override tick
        :type sqrt_price_x96: int
        :return: added position, base token used, quote token used
        :rtype: (PositionInfo, Decimal, Decimal)
        """
        if sqrt_price_x96 == -1 and tick != -1:
            sqrt_price_x96 = tick_to_sqrtPriceX96(tick)
        base_max_amount = self.base_asset.balance if base_max_amount is None else base_max_amount
        quote_max_amount = self.quote_asset.balance if quote_max_amount is None else quote_max_amount
        token0_amt, token1_amt = self.__convert_pair(base_max_amount, quote_max_amount)
        (created_position, token0_used, token1_used, liquidity) = self._add_liquidity_by_tick(token0_amt,
                                                                                              token1_amt,
                                                                                              lower_tick,
                                                                                              upper_tick,
                                                                                              sqrt_price_x96)
        base_used, quote_used = self.__convert_pair(token0_used, token1_used)
        self.action_buffer.append(
            AddLiquidityAction(base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                               quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                               base_amount_max=UnitDecimal(base_max_amount, self.base_asset.name),
                               quote_amount_max=UnitDecimal(quote_max_amount, self.quote_asset.name),
                               lower_quote_price=UnitDecimal(self.tick_to_price(lower_tick), self._price_unit),
                               upper_quote_price=UnitDecimal(self.tick_to_price(upper_tick), self._price_unit),
                               base_amount_actual=UnitDecimal(base_used, self.base_asset.name),
                               quote_amount_actual=UnitDecimal(quote_used, self.quote_asset.name),
                               position=created_position,
                               liquidity=int(liquidity)))
        return created_position, base_used, quote_used, liquidity

    @float_param_formatter
    def remove_liquidity(self, position: PositionInfo, liquidity: int = None, collect: bool = True,
                         sqrt_price_x96: int = -1, remove_dry_pool: bool = True) -> (Decimal, Decimal):
        """
        remove liquidity from pool, liquidity will be reduced to 0,
        instead of send tokens to broker, tokens will be transferred to fee property in position.
        position will be not deleted, until fees and tokens are collected.

        :param position: position to remove.
        :type position: PositionInfo
        :param liquidity: liquidity amount to remove, if set to None, all the liquidity will be removed
        :type liquidity: int
        :param collect: collect or not, if collect, will call collect function. and tokens will be sent to broker. if not token will be kept in fee property of postion
        :type collect: bool
        :param sqrt_price_x96: precise price.  if set to none, it will be calculated from current price.
        :type sqrt_price_x96: int
        :param remove_dry_pool: remove pool which liquidity==0, effect when collect==True
        :type remove_dry_pool: bool
        :return: (base_got,quote_get), base and quote token amounts collected from position
        :rtype:  (Decimal,Decimal)
        """
        if liquidity and liquidity < 0:
            raise DemeterError("liquidity should large than 0")
        token0_get, token1_get, delta_liquidity = self.__remove_liquidity(position, liquidity, sqrt_price_x96)

        base_get, quote_get = self.__convert_pair(token0_get, token1_get)
        self.action_buffer.append(
            RemoveLiquidityAction(
                base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                position=position,
                base_amount=UnitDecimal(base_get, self.base_asset.name),
                quote_amount=UnitDecimal(quote_get, self.quote_asset.name),
                removed_liquidity=delta_liquidity,
                remain_liquidity=self.positions[position].liquidity
            ))
        if collect:
            return self.collect_fee(position, remove_dry_pool)
        else:
            return base_get, quote_get

    @float_param_formatter
    def collect_fee(self,
                    position: PositionInfo,
                    max_collect_amount0: Decimal = None,
                    max_collect_amount1: Decimal = None,
                    remove_dry_pool: bool = True) -> (Decimal, Decimal):
        """
        collect fee and token from positions,
        if the amount and liquidity is zero, this position will be deleted.

        :param position: position to collect
        :type position: PositionInfo
        :param max_collect_amount0: max token0 amount to collect, eg: 1.2345 usdc, if set to None, all the amount will be collect
        :type max_collect_amount0: Decimal
        :param max_collect_amount1: max token0 amount to collect, if set to None, all the amount will be collect
        :type max_collect_amount1: Decimal
        :param remove_dry_pool: remove pool which liquidity==0, effect when collect==True
        :type remove_dry_pool: bool
        :return: (base_got,quote_get), base and quote token amounts collected from position
        :rtype:  (Decimal,Decimal)
        """
        if (max_collect_amount0 and max_collect_amount0 < 0) or \
                (max_collect_amount1 and max_collect_amount1 < 0):
            raise DemeterError("collect amount should large than 0")
        token0_get, token1_get = self.__collect_fee(self._positions[position])

        base_get, quote_get = self.__convert_pair(token0_get, token1_get)
        if self._positions[position]:
            self.action_buffer.append(
                CollectFeeAction(
                    base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                    quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                    position=position,
                    base_amount=UnitDecimal(base_get, self.base_asset.name),
                    quote_amount=UnitDecimal(quote_get, self.quote_asset.name)
                ))
        if self._positions[position].pending_amount0 == Decimal(0) \
                and self._positions[position].pending_amount1 == Decimal(0) \
                and self._positions[position].liquidity == 0 \
                and remove_dry_pool:
            del self.positions[position]
        return base_get, quote_get

    @float_param_formatter
    def buy(self, amount: Union[Decimal, float], price: Union[Decimal, float] = None) -> (Decimal, Decimal, Decimal):
        """
        buy token, swap from base token to quote token.

        :param amount: amount to buy(in quote token)
        :type amount:  Union[Decimal, float]
        :param price: price
        :type price: Union[Decimal, float]
        :return: fee, base token amount spend, quote token amount got
        :rtype: (Decimal, Decimal, Decimal)
        """
        price = price if price else self.pool_status.price
        from_amount = price * amount
        from_amount_with_fee = from_amount * (1 + self.pool_info.fee_rate)
        fee = from_amount_with_fee - from_amount
        from_asset, to_asset = self.__convert_pair(self._asset0, self._asset1)
        from_asset.sub(from_amount_with_fee)
        to_asset.add(amount)
        base_amount, quote_amount = self.__convert_pair(from_amount, amount)
        self.action_buffer.append(
            BuyAction(base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                      quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                      amount=UnitDecimal(amount, self.quote_asset.name),
                      price=UnitDecimal(price, self._price_unit),
                      fee=UnitDecimal(fee, self.base_asset.name),
                      base_change=UnitDecimal(base_amount, self.base_asset.name),
                      quote_change=UnitDecimal(quote_amount, self.quote_asset.name)))
        return fee, base_amount, quote_amount

    @float_param_formatter
    def sell(self, amount: Union[Decimal, float], price: Union[Decimal, float] = None) -> (Decimal, Decimal, Decimal):
        """
        sell token, swap from quote token to base token.

        :param amount: amount to sell(in quote token)
        :type amount:  Union[Decimal, float]
        :param price: price
        :type price: Union[Decimal, float]
        :return: fee, base token amount got, quote token amount spend
        :rtype: (Decimal, Decimal, Decimal)
        """
        price = price if price else self.pool_status.price
        from_amount_with_fee = amount
        from_amount = from_amount_with_fee * (1 - self.pool_info.fee_rate)
        to_amount = from_amount * price
        fee = from_amount_with_fee - from_amount
        to_asset, from_asset = self.__convert_pair(self._asset0, self._asset1)
        from_asset.sub(from_amount_with_fee)
        to_asset.add(to_amount)
        base_amount, quote_amount = self.__convert_pair(to_amount, from_amount)
        self.action_buffer.append(
            SellAction(base_balance_after=UnitDecimal(self.base_asset.balance, self.base_asset.name),
                       quote_balance_after=UnitDecimal(self.quote_asset.balance, self.quote_asset.name),
                       amount=UnitDecimal(amount, self.quote_asset.name),
                       price=UnitDecimal(price, self._price_unit),
                       fee=UnitDecimal(fee, self.quote_asset.name),
                       base_change=UnitDecimal(base_amount, self.base_asset.name),
                       quote_change=UnitDecimal(quote_amount, self.quote_asset.name)))

        return fee, base_amount, quote_amount

    def even_rebalance(self, price: Decimal = None) -> (Decimal, Decimal, Decimal):
        """
        Divide assets equally between two tokens.

        :param price: price of quote token. eg: 1234 eth/usdc
        :type price: Decimal
        :return: fee, base token amount spend, quote token amount got
        :rtype: (Decimal, Decimal, Decimal)
        """
        if price is None:
            price = self._pool_status.price

        total_capital = self.base_asset.balance + self.quote_asset.balance * price
        target_base_amount = total_capital / 2
        quote_amount_diff = target_base_amount / price - self.quote_asset.balance
        if quote_amount_diff > 0:
            return self.buy(quote_amount_diff)
        elif quote_amount_diff < 0:
            return self.sell(0 - quote_amount_diff)

    def remove_all_liquidity(self):
        """
        remove all the positions kept in broker.
        """
        if len(self.positions) < 1:
            return
        keys = list(self.positions.keys())
        for position_key in keys:
            self.remove_liquidity(position_key)



import datetime

import pandas
from pandas import Timestamp

from ._typing import MarketData, OnchainTxType, MarketDataNames
from .swap_contract import handle_event
from .utils import TextUtil, TimeUtil, DataUtil


class ModuleUtils(object):

    @staticmethod
    def get_datetime(date_str: str) -> datetime:
        if type(date_str) == Timestamp:
            return date_str.to_pydatetime()
        else:
            return datetime.datetime.strptime(TextUtil.cut_after(str(date_str), "+").replace("T", " "),
                                              "%Y-%m-%d %H:%M:%S")


def process_raw_data(raw_data: pandas.DataFrame) -> "pandas.DataFrame":
    if raw_data.size <= 0:
        return raw_data
    start_time = TimeUtil.get_minute(ModuleUtils.get_datetime(raw_data.loc[0, "block_timestamp"]))
    minute_rows = []
    data = []
    total_index = 1
    for index, row in raw_data.iterrows():
        current_time = TimeUtil.get_minute(ModuleUtils.get_datetime(row["block_timestamp"]))
        if start_time == current_time:  # middle of a minute
            minute_rows.append(row)
        else:  #
            data.append(sample_data_to_one_minute(start_time, minute_rows))
            total_index += 1
            # start on_bar minute
            start_time = current_time
            minute_rows = [row]
    data = DataUtil.fill_missing(data)
    df = pandas.DataFrame(columns=MarketDataNames, data=map(lambda d: d.to_array(), data))
    return df


def sample_data_to_one_minute(current_time, minute_rows) -> MarketData:
    data = MarketData()
    data.timestamp = current_time
    i = 1
    for r in minute_rows:
        tx_type, sender, receipt, amount0, amount1, sqrtPriceX96, current_liquidity, current_tick, tick_lower, tick_upper, delta_liquidity = handle_event(
            r.topics, r.DATA)
        # print(tx_type, sender, receipt, amount0, amount1, sqrtPriceX96, current_liquidity, current_tick, tick_lower,
        #       tick_upper, delta_liquidity)
        match tx_type:
            case OnchainTxType.MINT:
                pass
            case OnchainTxType.BURN:
                pass
            case OnchainTxType.COLLECT:
                pass
            case OnchainTxType.SWAP:
                data.netAmount0 += amount0
                data.netAmount1 += amount1
                if amount0 > 0:
                    data.inAmount0 += amount0
                if amount1 > 0:
                    data.inAmount1 += amount1
                if data.openTick is None:  # first
                    data.openTick = current_tick
                    data.highestTick = current_tick
                    data.lowestTick = current_tick
                if data.highestTick < current_tick:
                    data.highestTick = current_tick
                if data.lowestTick > current_tick:
                    data.lowestTick = current_tick
                if i == len(minute_rows):  # last
                    data.closeTick = current_tick
                    data.currentLiquidity = current_liquidity

        i += 1
    return data



from decimal import Decimal

# -*- coding: utf-8 -*-
"""
!!! IMPORTANT 

this module is developed and enhanced from active-strategy-framework of GammaStrategies
source code: https://github.com/GammaStrategies/active-strategy-framework/blob/main/UNI_v3_funcs.py

Original author information: 
=============================================

Created on Mon Jun 14 18:53:09 2021

@author: JNP
"""

'''liquitidymath'''
'''Python library to emulate the calculations done in liquiditymath.sol of UNI_V3 peryphery contract'''

# sqrtP: format X96 = int(1.0001**(tick/2)*(2**96))
# liquidity: int
# sqrtA = price for lower tick
# sqrtB = price for upper tick
'''get_amounts function'''


# Use 'get_amounts' function to calculate amounts as a function of liquitidy and price range
def get_amount0(sqrtA: int, sqrtB: int, liquidity: int, decimals: int) -> Decimal:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount0 = (Decimal(liquidity * 2 ** 96 * (sqrtB - sqrtA)) / sqrtB / sqrtA) / 10 ** decimals
    return amount0


def get_amount1(sqrtA: int, sqrtB: int, liquidity: int, decimals: int) -> Decimal:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount1 = Decimal(liquidity * (sqrtB - sqrtA)) / 2 ** 96 / 10 ** decimals
    return amount1


def get_sqrt(tick: int):
    return Decimal(1.0001 ** (tick / 2) * (2 ** 96))


def get_amounts(sqrt_price_x96: int, tickA: int, tickB: int, liquidity: int, decimal0: int, decimal1: int) -> \
        (Decimal, Decimal):
    sqrt = sqrt_price_x96
    sqrtA = get_sqrt_ratio_at_tick(tickA)
    sqrtB = get_sqrt_ratio_at_tick(tickB)

    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    if sqrt <= sqrtA:
        amount0 = get_amount0(sqrtA, sqrtB, liquidity, decimal0)
        return amount0, Decimal(0)
    elif sqrtB > sqrt > sqrtA:
        amount0 = get_amount0(sqrt, sqrtB, liquidity, decimal0)
        amount1 = get_amount1(sqrtA, sqrt, liquidity, decimal1)
        return amount0, amount1
    else:
        amount1 = get_amount1(sqrtA, sqrtB, liquidity, decimal1)
        return Decimal(0), amount1


'''get token amounts relation'''


# Use this formula to calculate amount of t0 based on amount of t1 (required before calculate liquidity)
# relation = t1/t0
def amounts_relation(tick: int, tickA: int, tickB: int, decimals0: int, decimals1: int) -> Decimal:
    sqrt = (1.0001 ** tick / 10 ** (decimals1 - decimals0)) ** (1 / 2)
    sqrtA = (1.0001 ** tickA / 10 ** (decimals1 - decimals0)) ** (1 / 2)
    sqrtB = (1.0001 ** tickB / 10 ** (decimals1 - decimals0)) ** (1 / 2)

    if sqrt == sqrtA or sqrt == sqrtB:
        relation = 0

    relation = (sqrt - sqrtA) / ((1 / sqrt) - (1 / sqrtB))
    return relation


'''get_liquidity function'''


def mul_div(a: int, b: int, denominator: int) -> int:
    """
    this function is very long in contract. but It's because max length in solidity is limited.

    But python has unlimit integer.

    ensure all the parameter is int !
    """
    return a * b // denominator


def get_liquidity_for_amount0(sqrtA: int, sqrtB: int, amount: int) -> int:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    intermediate = mul_div(sqrtA, sqrtB, 2 ** 96)
    return mul_div(amount, intermediate, sqrtB - sqrtA)


def get_liquidity_for_amount1(sqrtA: int, sqrtB: int, amount: int) -> int:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    return mul_div(amount, 2 ** 96, sqrtB - sqrtA)


def to_wei(amount, decimals) -> int:
    return int(amount * 10 ** decimals)


def get_liquidity(sqrt_price_x96: int, tickA: int, tickB: int,
                  amount0: Decimal, amount1: Decimal,
                  decimal0: int, decimal1: int) -> Decimal:
    sqrt = sqrt_price_x96
    sqrtA = get_sqrt_ratio_at_tick(tickA)
    sqrtB = get_sqrt_ratio_at_tick(tickB)

    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount0wei: int = to_wei(amount0, decimal0)
    amount1wei: int = to_wei(amount1, decimal1)
    if sqrt <= sqrtA:
        liquidity0 = get_liquidity_for_amount0(sqrtA, sqrtB, amount0wei)
        return liquidity0
    elif sqrtB > sqrt > sqrtA:
        liquidity0 = get_liquidity_for_amount0(sqrt, sqrtB, amount0wei)
        liquidity1 = get_liquidity_for_amount1(sqrtA, sqrt, amount1wei)
        liquidity = liquidity0 if liquidity0 < liquidity1 else liquidity1
        return liquidity
    else:
        liquidity1 = get_liquidity_for_amount1(sqrtA, sqrtB, amount1wei)
        return liquidity1


def get_sqrt_ratio_at_tick(tick: int) -> int:
    tick = int(tick)
    abs_tick = tick if tick >= 0 else -tick
    assert abs_tick <= 887272

    # 这些魔数分别表示 1/sqrt(1.0001)^1, 1/sqrt(1.0001)^2, 1/sqrt(1.0001)^4....
    ratio: int = 0xfffcb933bd6fad37aa2d162d1a594001 if abs_tick & 0x1 != 0 else 0x100000000000000000000000000000000
    if abs_tick & 0x2 != 0: ratio = (ratio * 0xfff97272373d413259a46990580e213a) >> 128
    if abs_tick & 0x4 != 0: ratio = (ratio * 0xfff2e50f5f656932ef12357cf3c7fdcc) >> 128
    if abs_tick & 0x8 != 0: ratio = (ratio * 0xffe5caca7e10e4e61c3624eaa0941cd0) >> 128
    if abs_tick & 0x10 != 0: ratio = (ratio * 0xffcb9843d60f6159c9db58835c926644) >> 128
    if abs_tick & 0x20 != 0: ratio = (ratio * 0xff973b41fa98c081472e6896dfb254c0) >> 128
    if abs_tick & 0x40 != 0: ratio = (ratio * 0xff2ea16466c96a3843ec78b326b52861) >> 128
    if abs_tick & 0x80 != 0: ratio = (ratio * 0xfe5dee046a99a2a811c461f1969c3053) >> 128
    if abs_tick & 0x100 != 0: ratio = (ratio * 0xfcbe86c7900a88aedcffc83b479aa3a4) >> 128
    if abs_tick & 0x200 != 0: ratio = (ratio * 0xf987a7253ac413176f2b074cf7815e54) >> 128
    if abs_tick & 0x400 != 0: ratio = (ratio * 0xf3392b0822b70005940c7a398e4b70f3) >> 128
    if abs_tick & 0x800 != 0: ratio = (ratio * 0xe7159475a2c29b7443b29c7fa6e889d9) >> 128
    if abs_tick & 0x1000 != 0: ratio = (ratio * 0xd097f3bdfd2022b8845ad8f792aa5825) >> 128
    if abs_tick & 0x2000 != 0: ratio = (ratio * 0xa9f746462d870fdf8a65dc1f90e061e5) >> 128
    if abs_tick & 0x4000 != 0: ratio = (ratio * 0x70d869a156d2a1b890bb3df62baf32f7) >> 128
    if abs_tick & 0x8000 != 0: ratio = (ratio * 0x31be135f97d08fd981231505542fcfa6) >> 128
    if abs_tick & 0x10000 != 0: ratio = (ratio * 0x9aa508b5b7a84e1c677de54f3e99bc9) >> 128
    if abs_tick & 0x20000 != 0: ratio = (ratio * 0x5d6af8dedb81196699c329225ee604) >> 128
    if abs_tick & 0x40000 != 0: ratio = (ratio * 0x2216e584f5fa1ea926041bedfe98) >> 128
    if abs_tick & 0x80000 != 0: ratio = (ratio * 0x48a170391f7dc42444e8fa2) >> 128

    if tick > 0:
        # type(uint256).max
        ratio = int(115792089237316195423570985008687907853269984665640564039457584007913129639935 // ratio)

    # this divides by 1<<32 rounding up to go from a Q128.128 to a Q128.96.
    # we then downcast because we know the result always fits within 160 bits due to our tick input constraint
    # we round up in the division so getTickAtSqrtRatio of the output price is always consistent
    sqrt_price_x96 = (ratio >> 32) + (0 if ratio % (1 << 32) == 0 else 1)
    return sqrt_price_x96



from decimal import Decimal

from .helper import quote_price_to_tick, from_wei
from .liquitidymath import get_amounts, get_liquidity
from .types import PoolBaseInfo, Position, PoolStatus
from .._typing import PositionInfo


class V3CoreLib(object):
    @staticmethod
    def new_position(pool: PoolBaseInfo,
                     token0_amount: Decimal, token1_amount: Decimal,
                     lower_tick: int, upper_tick: int, sqrt_price_x96: int):
        position_liq = get_liquidity(sqrt_price_x96, lower_tick, upper_tick,
                                     token0_amount, token1_amount,
                                     pool.token0.decimal, pool.token1.decimal)
        token0_in_position, token1_in_position = get_amounts(sqrt_price_x96, lower_tick, upper_tick,
                                                             position_liq, pool.token0.decimal, pool.token1.decimal)
        new_position_entity = PositionInfo(lower_tick=lower_tick,
                                           upper_tick=upper_tick)
        return token0_in_position, token1_in_position, int(position_liq), new_position_entity

    @staticmethod
    def close_position(pool: PoolBaseInfo, position_info: PositionInfo, liquidity, sqrt_price_x96):
        return V3CoreLib.get_token_amounts(pool, position_info, sqrt_price_x96, liquidity)

    @staticmethod
    def get_token_amounts(pool: PoolBaseInfo, pos: PositionInfo, sqrt_price_x96, liquidity) -> (Decimal, Decimal):
        amount0, amount1 = get_amounts(sqrt_price_x96,
                                       pos.lower_tick,
                                       pos.upper_tick,
                                       liquidity,
                                       pool.token0.decimal,
                                       pool.token1.decimal)
        return amount0, amount1

    @staticmethod
    def quote_price_pair_to_tick(pool: PoolBaseInfo, lower_quote_price: Decimal, upper_quote_price: Decimal):
        lower_tick = quote_price_to_tick(lower_quote_price, pool.token0.decimal, pool.token1.decimal,
                                         pool.is_token0_base)
        upper_tick = quote_price_to_tick(upper_quote_price, pool.token0.decimal, pool.token1.decimal,
                                         pool.is_token0_base)
        return lower_tick, upper_tick

    @staticmethod
    def update_fee(pool: PoolBaseInfo, pos: PositionInfo, position: Position, state: PoolStatus):
        # in most cases, tick will not cross to on_bar one, which means L will not change.
        if pos.upper_tick > state.current_tick > pos.lower_tick:
            # if the simulating liquidity is above the actual liquidity, we will consider share=1
            if position.liquidity >= state.current_liquidity:
                share = Decimal(1)
            else:
                share = Decimal(position.liquidity) / Decimal(state.current_liquidity)
            position.pending_amount0 += from_wei(state.in_amount0, pool.token0.decimal) * share * pool.fee_rate
            position.pending_amount1 += from_wei(state.in_amount1, pool.token1.decimal) * share * pool.fee_rate



from pathlib import Path

from setuptools import setup, find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='zelos-demeter',
    version='0.1.8',
    packages=find_packages(exclude=["tests", "tests.*", "samples", "samples.*"]),
    url='https://zelos-demeter.readthedocs.io',
    license='MIT',
    author='zelos research',
    author_email='zelos@antalpha.com',
    description='better back testing tool for uniswap v3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.10',
    install_requires=["pandas>=1.4.4",
                      "tqdm>=4.64.1",
                      "google-cloud-bigquery>=3.3.5",
                      "db-dtypes>=1.0.4",
                      "toml>=0.10.2"],
)

# rm -rf ./demeter.egg-info/ && python setup.py sdist upload -r private
# python setup.py sdist
# twine upload -r pypi dist/*



from decimal import Decimal
from typing import Union

import pandas as pd

from .trigger import Trigger
from .._typing import PositionInfo, BaseAction, AddLiquidityAction, SellAction, BuyAction, CollectFeeAction, \
    RemoveLiquidityAction, RowData
from ..broker import Broker
from ..data_line import Lines, Line


class Strategy(object):
    """
    strategy parent class, all user strategy should inherit this class
    """

    def __init__(self):
        self.broker: Broker = None
        self.data: Lines = None
        self.number_format = ".8g"
        self.triggers: [Trigger] = []

    def initialize(self):
        """
        initialize your strategy, this will be called before self.on_bar()

        """
        pass

    def before_bar(self, row_data: Union[RowData, pd.Series]):
        """
        called before triggers on each row, at this time, fees are not updated yet. you can add some indicator or add some actions

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[RowData, pd.Series]
        """
        pass

    def on_bar(self, row_data: Union[RowData, pd.Series]):
        """
        called after triggers on each row, at this time, fees and account status are not updated yet. you can add some actions here

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[RowData, pd.Series]
        """
        pass

    def after_bar(self, row_data: Union[RowData, pd.Series]):
        """
        called after fees and account status are updated on each row. you can add some statistic logic here

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[RowData, pd.Series]
        """
        pass

    def finalize(self):
        """
        this will run after all the data processed.

        """
        pass

    def notify(self, action: BaseAction):
        """
        notify if non-basic action happens

        :param action:  action
        :type action: BaseAction
        """
        print(action.get_output_str())

    def notify_add_liquidity(self, action: AddLiquidityAction):
        """
        notify if add liquidity action happens

        :param action:  action
        :type action: AddLiquidityAction
        """
        print(action.get_output_str())

    def notify_remove_liquidity(self, action: RemoveLiquidityAction):
        """
        notify if remove liquidity action happens

        :param action:  action
        :type action: RemoveLiquidityAction
        """
        print(action.get_output_str())

    def notify_collect_fee(self, action: CollectFeeAction):
        """
        notify if collect fee action happens

        :param action:  action
        :type action: CollectFeeAction
        """
        print(action.get_output_str())

    def notify_buy(self, action: BuyAction):
        """
        notify if buy action happens

        :param action:  action
        :type action: BuyAction
        """
        print(action.get_output_str())

    def notify_sell(self, action: SellAction):
        """
        notify if sell action happens

        :param action:  action
        :type action: SellAction
        """
        print(action.get_output_str())

    def _add_column(self, name: str, line: Line):
        """
        add a column to data

        :param name: column name
        :type name: str
        :param line: data
        :type line: Line
        """
        self.data[name] = line

    def add_liquidity(self,
                      lower_quote_price: Union[Decimal, float],
                      upper_quote_price: Union[Decimal, float],
                      base_max_amount: Union[Decimal, float] = None,
                      quote_max_amount: Union[Decimal, float] = None,
                      ) -> (PositionInfo, Decimal, Decimal):
        """

        add liquidity, then get a new position

        :param lower_quote_price: lower price base on quote token.
        :type lower_quote_price: Union[Decimal, float]
        :param upper_quote_price: upper price base on quote token.
        :type upper_quote_price: Union[Decimal, float]
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Union[Decimal, float]
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Union[Decimal, float]
        :return: added position, base token used, quote token used
        :rtype: (PositionInfo, Decimal, Decimal)
        """

        return self.broker.add_liquidity(lower_quote_price, upper_quote_price, base_max_amount, quote_max_amount)

    def add_liquidity_by_tick(self,
                              lower_tick: int,
                              upper_tick: int,
                              base_max_amount: Union[Decimal, float] = None,
                              quote_max_amount: Union[Decimal, float] = None,
                              sqrt_price_x96: int = -1,
                              tick: int = -1):
        """

        add liquidity, you need to set tick instead of price.

        :param lower_tick: lower tick
        :type lower_tick: int
        :param upper_tick: upper tick
        :type upper_tick: int
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Union[Decimal, float]
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Union[Decimal, float]
        :param tick: tick price.  if set to none, it will be calculated from current price.
        :type tick: int
        :param sqrt_price_x96: precise price.  if set to none, it will be calculated from current price. this param will override tick
        :type sqrt_price_x96: int
        :return: added position, base token used, quote token used
        :rtype: (PositionInfo, Decimal, Decimal)
        """
        return self.broker.add_liquidity_by_tick(lower_tick,
                                                 upper_tick,
                                                 base_max_amount,
                                                 quote_max_amount,
                                                 sqrt_price_x96,
                                                 tick)

    def remove_liquidity(self, positions: Union[PositionInfo, list], remove_dry_pool: bool = True) -> {
        PositionInfo: (Decimal, Decimal)}:
        """
        remove liquidity from pool, position will be deleted

        :param positions: position info, as an object or an array
        :type positions: [PositionInfo]
        :return: a dict, key is position info, value is (base_got,quote_get), base_got is base token amount collected from position
        :rtype: {PositionInfo: (Decimal,Decimal)}
        """
        return self.broker.remove_liquidity(positions, remove_dry_pool)

    def collect_fee(self, positions: [PositionInfo]) -> {PositionInfo: tuple}:
        """
        collect fee from positions

        :param positions: position info, as an object or an array
        :type positions: [PositionInfo]
        :return: a dict, key is position info, value is (base_got,quote_get), base_got is base token fee collected from position
        :rtype: {Position: tuple(base_got,quote_get)}
        """
        return self.broker.collect_fee(positions)

    def buy(self, amount: Union[Decimal, float], price: Union[Decimal, float] = None) -> (Decimal, Decimal, Decimal):
        """
        buy token, swap from base token to quote token.

        :param amount: amount to buy(in quote token)
        :type amount:  Union[Decimal, float]
        :param price: price
        :type price: Union[Decimal, float]
        :return: fee, base token amount spend, quote token amount got
        :rtype: (Decimal, Decimal, Decimal)
        """
        return self.broker.buy(amount, price)

    def sell(self, amount: Union[Decimal, float], price: Union[Decimal, float] = None) -> (Decimal, Decimal, Decimal):
        """
        sell token, swap from quote token to base token.

        :param amount: amount to sell(in quote token)
        :type amount:  Union[Decimal, float]
        :param price: price
        :type price: Union[Decimal, float]
        :return: fee, base token amount got, quote token amount spend
        :rtype: (Decimal, Decimal, Decimal)
        """
        return self.broker.sell(amount, price)



import os
import sys
from datetime import datetime

import toml

from .download import ChainType, DataSource, downloader, DownloadParam
from .utils.application import get_enum_by_name, dict_to_object


class Downloader:

    def __init__(self, config):
        self.config: DownloadParam = self.convert_config(config)

    @staticmethod
    def convert_config(config: dict):
        param = DownloadParam()
        if hasattr(config, "chain"):
            param.chain = get_enum_by_name(ChainType, config.chain)
        if hasattr(config, "source"):
            param.source = get_enum_by_name(DataSource, config.source)
        if hasattr(config, "save_path"):
            param.save_path = config.save_path
        if not param.save_path.endswith("/"):
            param.save_path += "/"
        if not os.path.exists(param.save_path):
            os.makedirs(param.save_path)
        if hasattr(config, "pool_address"):
            param.pool_address = config.pool_address
        else:
            raise RuntimeError("you must set pool contract address")
        if param.source == DataSource.BigQuery:
            if not hasattr(config, "big_query"):
                raise RuntimeError("must set big_query parameters")
            if hasattr(config.big_query, "auth_file"):
                param.big_query.auth_file = config.big_query.auth_file
                if not os.path.exists(param.big_query.auth_file):
                    raise RuntimeError("google auth file not found")
            else:
                raise RuntimeError("you must set google auth file")
            if hasattr(config.big_query, "start"):
                param.big_query.start = config.big_query.start
            else:
                raise RuntimeError("you must set start date, eg: 2022-10-1")
            if hasattr(config.big_query, "end"):
                param.big_query.end = config.big_query.end
        elif param.source == DataSource.RPC:
            if not hasattr(config, "rpc"):
                raise RuntimeError("must set rpc parameters")
            if hasattr(config.rpc, "end_point"):
                param.rpc.end_point = config.rpc.end_point
            else:
                raise RuntimeError("you must set end_point")
            if hasattr(config.rpc, "start_height"):
                param.rpc.start_height = config.rpc.start_height
            else:
                raise RuntimeError("you must set start_height")
            if hasattr(config.rpc, "end_height"):
                param.rpc.end_height = config.rpc.end_height
            else:
                raise RuntimeError("you must set end_height")
            if hasattr(config.rpc, "auth_string"):
                param.rpc.auth_string = config.rpc.auth_string
            if hasattr(config.rpc, "proxy"):
                param.rpc.proxy = config.rpc.proxy
            if hasattr(config.rpc, "batch_size"):
                param.rpc.batch_size = config.rpc.batch_size
        return param

    def do_download(self):
        if self.config.source == DataSource.BigQuery:
            start_date = datetime.strptime(self.config.big_query.start, "%Y-%m-%d").date()
            end_date = datetime.strptime(self.config.big_query.end, "%Y-%m-%d").date()
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self.config.big_query.auth_file
            downloader.download_from_bigquery(self.config.chain,
                                              self.config.pool_address,
                                              start_date,
                                              end_date,
                                              self.config.save_path,
                                              save_raw_file=True)
        elif self.config.source == DataSource.RPC:
            downloader.download_from_rpc(self.config)
        print("download complete, check your files in " + self.config.save_path)


if __name__ == '__main__':
    if len(sys.argv) == 1:
        print("please set a config file. in toml format. eg: 'python -m demeter.downloader config.toml'.")
        exit(1)
    if not os.path.exists(sys.argv[1]):
        print("config file not found,")
        exit(1)
    config_file = toml.load(sys.argv[1])
    try:
        download_entity = Downloader(dict_to_object(config_file))
    except RuntimeError as e:
        print(e)
        exit(1)
    print(download_entity.config)
    download_entity.do_download()



import os
from datetime import date, timedelta
import json
import pandas as pd
from tqdm import tqdm  # process bar

from ._typing import ChainType, DownloadParam
from .process import process_raw_data
from .source_bigquery import download_bigquery_pool_event_oneday
from .source_rpc import download_and_save_by_day
from .utils import get_file_name


def download_from_rpc(config: DownloadParam):
    if config.rpc.end_height <= config.rpc.start_height:
        raise RuntimeError("start height should less than end height")

    downloaded_day = download_and_save_by_day(config)
    # downloaded_day = ["2022-06-30", "2022-07-01", "2022-07-02"]

    if len(downloaded_day) <= 2:
        raise RuntimeError("As first day and last day will be dropped, "
                           "day length should at least 3, current length is " + len(downloaded_day))
    print(f"now will drop data in {downloaded_day[0]} and {downloaded_day[len(downloaded_day) - 1]} "
          f"as they a highly likely insufficient for a whole day")
    downloaded_day = downloaded_day[1:len(downloaded_day) - 1]
    for day in downloaded_day:
        day_df = pd.read_csv(get_file_name(config.save_path, config.chain.name, config.pool_address, day, True))
        processed_day_data = process_raw_data(day_df)
        processed_day_data.to_csv(get_file_name(config.save_path, config.chain.name, config.pool_address, day, False),
                                  header=True, index=False)


def download_from_bigquery(chain: ChainType, pool_address: str, start: date, end: date,
                           save_path=os.getcwd(), save_raw_file=False, skip_exist=True):
    """
    Download transfer data by day
    :param chain: which chain
    :param pool_address: contract address of swap pool
    :param start: start date
    :param end: end date
    :param data_source: which data source to download
    :param save_path: save to path
    :param save_raw_file: save raw data or not
    :param skip_exist: if file exist, skip.
    :return:
    """
    pool_address = pool_address.lower()
    end = end + timedelta(days=1)  # make date range is [a,b], instead of [a,b)
    if start > end:
        raise RuntimeError("start date should earlier than end date")
    date_array = split_date_range_to_array(start, end)
    for i in tqdm(range(len(date_array)), ncols=150):
        day = date_array[i]
        date_str = day.strftime("%Y-%m-%d")
        file_name = get_file_name(save_path, chain.name, pool_address, date_str, False)
        if skip_exist and os.path.exists(file_name):
            continue
        raw_day_data = download_bigquery_pool_event_oneday(chain, pool_address, day)
        if save_raw_file:
            raw_day_data['topics'] = raw_day_data['topics'].apply(lambda x: json.dumps(x.tolist()))
            raw_day_data.to_csv(get_file_name(save_path, chain.name, pool_address, date_str, True),
                                header=True,
                                index=False)
        processed_day_data = process_raw_data(raw_day_data)
        # save processed
        processed_day_data.to_csv(file_name, header=True, index=False)
        # time.sleep(1)


def split_date_range_to_array(start: date, end: date) -> "array":
    return [start + timedelta(days=x) for x in range(0, (end - start).days)]



import logging
from datetime import date, timedelta, datetime
from decimal import Decimal

import pandas as pd
from tqdm import tqdm  # process bar

from .evaluating_indicator import Evaluator
from .. import PoolStatus
from .._typing import AccountStatus, BarStatusNames, BaseAction, Asset, DemeterError, ActionTypeEnum, \
    EvaluatingIndicator, RowData
from ..broker import Broker, PoolBaseInfo
from ..data_line import Lines
from ..strategy import Strategy

DEFAULT_DATA_PATH = "./data"


def decimal_from_value(value):
    return Decimal(value)


class Actuator(object):
    """
    Core component of a back test. Manage the resources in a test, including broker/strategy/data/indicator,

    :param pool_info: pool information
    :type pool_info: PoolBaseInfo

    """

    def __init__(self, pool_info: PoolBaseInfo):
        self._broker: Broker = Broker(pool_info)
        # data
        self._data: Lines = None
        # strategy
        self._strategy: Strategy = Strategy()
        # all the actions during the test(buy/sell/add liquidity)
        self._actions: [BaseAction] = []
        # actions in current bar
        self.bar_actions: [BaseAction] = []
        self._broker.action_buffer = self.bar_actions
        # broker status in every bar, use array for performance
        self.account_status_list: [AccountStatus] = []
        # path of source data, which is saved by downloader
        self._data_path: str = DEFAULT_DATA_PATH
        # evaluating indicator calculator
        self._evaluator: Evaluator = None
        self._enable_evaluating: bool = True
        # logging
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger(__name__)

        # internal var
        self.__backtest_finished = False

    @property
    def account_status(self) -> pd.DataFrame:
        index = self.data.index[0:len(self.account_status_list)]
        return pd.DataFrame(columns=BarStatusNames,
                            index=index,
                            data=map(lambda d: d.to_array(), self.account_status_list))

    @property
    def final_status(self) -> AccountStatus:
        """
        Get status after back test finish.

        If test has not run, an error will be raised.

        :return: Final state of broker
        :rtype: AccountStatus
        """
        if self.__backtest_finished:
            return self.account_status_list[len(self.account_status_list) - 1]
        else:
            raise DemeterError("please run strategy first")

    def reset(self):
        """

        reset all the status variables

        """
        self._actions = []
        self._evaluator = None
        self.account_status_list = []
        self.__backtest_finished = False
        self.data.reset_cursor()

    @property
    def data_path(self) -> str:
        """
        path of source data, which is saved by downloader
        """
        return self._data_path

    @data_path.setter
    def data_path(self, value: str):
        """
        path of source data, which is saved by downloader

        :param value: path
        :type value: str

        """
        self._data_path = value

    @property
    def actions(self) -> [BaseAction]:
        """
        all the actions during the test(buy/sell/add liquidity)

        :return: action list
        :rtype: [BaseAction]
        """
        return self._actions

    @property
    def evaluating_indicator(self) -> EvaluatingIndicator:
        """
        evaluating indicator result

        :return:  evaluating indicator
        :rtype: EvaluatingIndicator
        """
        return self._evaluator.evaluating_indicator if self._evaluator is not None else None

    @property
    def broker(self) -> Broker:
        """
        Broker manage assets in back testing. Including asset, positions. it also provides operations for positions,

        :return: Broker
        :rtype: Broker
        """
        return self._broker

    @property
    def data(self) -> Lines:
        """
        data

        :return:
        :rtype: Lines
        """
        return self._data

    @data.setter
    def data(self, value: Lines):
        """
        set data which saved by downloader. data source should contain the following column, and indexed with timestamp

        * netAmount0,
        * netAmount1,
        * closeTick,
        * openTick,
        * lowestTick,
        * highestTick,
        * inAmount0,
        * inAmount1,
        * currentLiquidity

        :param value: data
        :type value: Lines
        """
        self._data = value

    @property
    def strategy(self) -> Strategy:
        """
        strategy,

        :return: strategy
        :rtype: Strategy
        """
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        """
        set strategy
        :param value: strategy
        :type value: Strategy
        """
        self._strategy = value

    @property
    def number_format(self) -> str:
        """
        number format for console output, eg: ".8g", ".5f"

        :return: number format
        :rtype: str
        """
        return self._number_format

    @number_format.setter
    def number_format(self, value: str):
        """
        number format for console output, eg: ".8g", ".5f", follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html

        :param value: number format,
        :type value:str
        """
        self._number_format = value

    def set_assets(self, assets=[Asset]):
        """
        set initial balance for token

        :param assets: assets to set.
        :type assets: [Asset]
        """
        for asset in assets:
            self._broker.set_asset(asset.token, asset.amount)

    def notify(self, strategy: Strategy, actions: [BaseAction]):
        """

        notify user when new action happens.

        :param strategy: Strategy
        :type strategy: Strategy
        :param actions:  action list
        :type actions: [BaseAction]
        """
        if len(actions) < 1:
            return
        last_time = datetime(1970, 1, 1)
        for action in actions:
            if last_time != action.timestamp:
                print(f"\033[7;34m{action.timestamp} \033[0m")
                last_time = action.timestamp
            match action.action_type:
                case ActionTypeEnum.add_liquidity:
                    strategy.notify_add_liquidity(action)
                case ActionTypeEnum.remove_liquidity:
                    strategy.notify_remove_liquidity(action)
                case ActionTypeEnum.collect_fee:
                    strategy.notify_collect_fee(action)
                case ActionTypeEnum.buy:
                    strategy.notify_buy(action)
                case ActionTypeEnum.sell:
                    strategy.notify_sell(action)
                case _:
                    strategy.notify(action)

    def load_data(self, chain: str, contract_addr: str, start_date: date, end_date: date):
        """

        load data, and preprocess. preprocess actions including:

        * fill empty data
        * calculate statistic column
        * set timestamp as index

        :param chain: chain name
        :type chain: str
        :param contract_addr: pool contract address
        :type contract_addr: str
        :param start_date: start test date
        :type start_date: date
        :param end_date: end test date
        :type end_date: date
        """
        self.logger.info(f"start load files from {start_date} to {end_date}...")
        df = pd.DataFrame()
        day = start_date
        while day <= end_date:
            path = f"{self.data_path}/{chain}-{contract_addr}-{day.strftime('%Y-%m-%d')}.csv"
            day_df = pd.read_csv(path, converters={'inAmount0': decimal_from_value,
                                                   'inAmount1': decimal_from_value,
                                                   'netAmount0': decimal_from_value,
                                                   'netAmount1': decimal_from_value,
                                                   "currentLiquidity": decimal_from_value})
            df = pd.concat([df, day_df])
            day = day + timedelta(days=1)
        self.logger.info("load file complete, preparing...")

        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)

        # fill empty row (first minutes in a day, might be blank)
        full_indexes = pd.date_range(start=df.index[0], end=df.index[df.index.size - 1], freq="1min")
        df = df.reindex(full_indexes)
        df = Lines.from_dataframe(df)
        df = df.fillna()
        self.add_statistic_column(df)
        self.data = df
        self.logger.info("data has benn prepared")

    def add_statistic_column(self, df: Lines):
        """
        add statistic column to data, new columns including:

        * open: open price
        * price: close price (current price)
        * low: lowest price
        * high: height price
        * volume0: swap volume for token 0
        * volume1: swap volume for token 1

        :param df: original data
        :type df: Lines

        """
        # add statistic column
        df["open"] = df["openTick"].map(lambda x: self.broker.tick_to_price(x))
        df["price"] = df["closeTick"].map(lambda x: self.broker.tick_to_price(x))
        high_name, low_name = ("lowestTick", "highestTick") if self.broker.pool_info.is_token0_base \
            else ("highestTick", "lowestTick")
        df["low"] = df[high_name].map(lambda x: self.broker.tick_to_price(x))
        df["high"] = df[low_name].map(lambda x: self.broker.tick_to_price(x))
        df["volume0"] = df["inAmount0"].map(lambda x: Decimal(x) / 10 ** self.broker.pool_info.token0.decimal)
        df["volume1"] = df["inAmount1"].map(lambda x: Decimal(x) / 10 ** self.broker.pool_info.token1.decimal)

    def run(self, enable_notify=True, enable_evaluating=True, print_final_status=False):
        """
        start back test, the whole process including:

        * reset actuator
        * initialize strategy (set object to strategy, then run strategy.initialize())
        * process each bar in data
            * prepare data in each row
            * run strategy.on_bar()
            * calculate fee earned
            * get latest account status
            * notify actions
        * run evaluator indicator
        * run strategy.finalize()

        :param enable_notify: notify when new action happens
        :type enable_notify: bool
        :param enable_evaluating: enable evaluating indicator. if not enabled, no evaluating will be calculated
        :type enable_evaluating: bool
        :param print_final_status: enable output.
        :type print_final_status: bool
        """
        self._enable_evaluating = enable_evaluating
        self.reset()
        if self._data is None:
            return
        self.logger.info("init strategy...")
        first_data = self._data.iloc[0]
        self._broker.pool_status = PoolStatus(self._data.index[0].to_pydatetime(),
                                              int(first_data.closeTick),
                                              first_data.currentLiquidity,
                                              first_data.inAmount0,
                                              first_data.inAmount1,
                                              first_data.price)
        self.init_strategy()
        if not isinstance(self._data, Lines):
            raise DemeterError("Data must be instance of Lines")
        row_id = 0
        first = True
        self.logger.info("start main loop...")
        with tqdm(total=len(self._data.index), ncols=150) as pbar:
            for index, row in self._data.iterrows():
                row_data = RowData()
                setattr(row_data, "timestamp", index.to_pydatetime())
                setattr(row_data, "row_id", row_id)
                row_id += 1
                for column_name in row.index:
                    setattr(row_data, column_name, row[column_name])
                # execute strategy, and some calculate
                # update price tick
                self._broker.pool_status = PoolStatus(index.to_pydatetime(),
                                                      int(row_data.closeTick),
                                                      row_data.currentLiquidity,
                                                      row_data.inAmount0,
                                                      row_data.inAmount1,
                                                      row_data.price)
                self._strategy.before_bar(row_data)

                if self._strategy.triggers:
                    for trigger in self._strategy.triggers:
                        if trigger.when(row_data):
                            trigger.do(row_data)
                self._strategy.on_bar(row_data)

                # update broker status, eg: re-calculate fee
                # and read the latest status from broker
                self._broker.update()

                self._strategy.after_bar(row_data)

                if first:
                    init_price = row_data.price
                    first = False
                self.account_status_list.append(self._broker.get_account_status(row_data.price, index.to_pydatetime()))

                # collect actions in this loop
                current_event_list = self.bar_actions.copy()
                for event in current_event_list:
                    event.timestamp = index
                self.bar_actions.clear()
                if current_event_list and len(current_event_list) > 0:
                    self._actions.extend(current_event_list)

                # process on_bar
                self._data.move_cursor_to_next()
                pbar.update()
        # notify
        if enable_notify and len(self._actions) > 0:
            self.logger.info("start notify all the actions")
            self.notify(self.strategy, self._actions)
        self.logger.info("main loop finished, start calculate evaluating indicator...")
        bar_status_df = pd.DataFrame(columns=BarStatusNames,
                                     index=self.data.index,
                                     data=map(lambda d: d.to_array(), self.account_status_list))
        self.logger.info("run evaluating indicator")
        if self._enable_evaluating:
            self._evaluator = Evaluator(
                self._broker.get_init_account_status(init_price, self.data.index[0].to_pydatetime()),
                bar_status_df)
            self._evaluator.run()
        self._strategy.finalize()
        self.logger.info("back testing finish")
        self.__backtest_finished = True
        if print_final_status:
            self.output()

    def output(self):
        """
        output back test result to console
        """
        if self.__backtest_finished:
            print("Final status")
            print(self.broker.get_account_status(self.data.tail(1).price[0]).get_output_str())
            if self._enable_evaluating:
                print("Evaluating indicator")
                print(self._evaluator.evaluating_indicator.get_output_str())
        else:
            raise DemeterError("please run strategy first")

    def init_strategy(self):
        """
        initialize strategy, set property to strategy. and run strategy.initialize()
        """
        if not isinstance(self._strategy, Strategy):
            raise DemeterError("strategy must be inherit from Strategy")
        self._strategy.broker = self._broker
        self._strategy.data = self._data
        self._strategy.account_status = self.account_status
        self._strategy.initialize()

    def __str__(self):
        return f"Demeter(broker:{self._broker})\n" \
               f"data(data length: {0 if self._data is None else len(self._data)})"



from ._typing import UnitDecimal, TokenInfo, Asset, ActionTypeEnum, PositionInfo, AccountStatus, RowData, BaseAction, \
    AddLiquidityAction, CollectFeeAction, RemoveLiquidityAction, BuyAction, SellAction, EvaluatingIndicator, \
    DemeterError, TimeUnitEnum, EthError
from .broker import Broker, tick_to_quote_price, PoolBaseInfo, PoolStatus, BrokerAsset, Position
from .core import Actuator
from .data_line import LineTypeEnum, Cursorable, Line, Lines
from .download import ChainType, DataSource
from .indicator import simple_moving_average, exponential_moving_average, actual_volatility
from .strategy import Strategy, Trigger, TimeRangesTrigger, TimeRangeTrigger, TimeRange, PeriodTrigger, PeriodsTrigger, \
    AtTimesTrigger, AtTimeTrigger



from datetime import datetime, timedelta

from ._typing import MarketData


class TextUtil(object):
    @staticmethod
    def cut_after(text: str, symbol: str) -> 'str':
        index = text.find(symbol)
        return text[0:index]


class TimeUtil(object):
    @staticmethod
    def get_minute(time: datetime) -> datetime:
        return datetime(time.year, time.month, time.day, time.hour, time.minute, 0)


class HexUtil(object):
    @staticmethod
    def to_signed_int(h):
        """
        Converts hex values to signed integers.
        """
        s = bytes.fromhex(h[2:])
        i = int.from_bytes(s, 'big', signed=True)
        return i


def get_file_name(path, chain, addr, day, raw: bool):
    raw_str = "raw-" if raw else ""
    return f"{path}{raw_str}{chain}-{addr.lower()}-{day}.csv"


class DataUtil(object):

    @staticmethod
    def fill_missing(data_list: [MarketData]) -> list:
        if len(data_list) < 1:
            return data_list
        # take the first minute in data. instead of 0:00:00
        # so here will be a problem, if the first data is 0:03:00, the first 2 minutes will be blank
        # that's because there is no previous data to follow
        # those empty rows will be filled in loading stage
        index_minute = data_list[0].timestamp
        new_list = []
        data_list_index = 0

        start_day = data_list[0].timestamp.day
        while index_minute.day == start_day:
            if (data_list_index < len(data_list)) and (index_minute == data_list[data_list_index].timestamp):
                item = data_list[data_list_index]
                data_list_index += 1
            else:
                item = MarketData()
                item.timestamp = index_minute
            prev_data = new_list[len(new_list) - 1] if len(new_list) - 1 >= 0 else None
            # if no previous(this might happen in the first minutes) data, this row will be discarded
            if item.fill_missing_field(prev_data):
                new_list.append(item)
            index_minute = index_minute + timedelta(minutes=1)

        return new_list



import datetime

import requests
from dataclasses import dataclass
import random

from .._typing import EthError


@dataclass
class GetLogsParam:
    address: str
    fromBlock: int
    toBlock: int
    topics: list[str]


class EthRequestClient:
    def __init__(self, endpoint: str, proxy="", auth=""):
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(pool_connections=5, pool_maxsize=20)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)
        self.headers = {}
        self.endpoint = endpoint
        if auth:
            self.headers["Authorization"] = auth
        self.proxies = {"http": proxy, "https": proxy, } if proxy else {}

    def __del__(self):
        self.session.close()

    @staticmethod
    def __encode_json_rpc(method: str, params: list):
        return {"jsonrpc": "2.0", "method": method, "params": params, "id": random.randint(1, 2147483648)}

    @staticmethod
    def __decode_json_rpc(response: requests.Response):
        content = response.json()
        if "error" in content:
            raise EthError(content["error"]["code"], content["error"]["message"])
        return content["result"]

    def do_post(self, param):
        return self.session.post(self.endpoint,
                                 json=param,
                                 proxies=self.proxies,
                                 headers=self.headers)

    def get_block(self, height):
        response = self.do_post(EthRequestClient.__encode_json_rpc("eth_getBlockByNumber", [hex(height), False]))
        return EthRequestClient.__decode_json_rpc(response)

    def get_block_timestamp(self, height):
        resp = self.get_block(height)
        if resp:
            timestamp = int(resp["timestamp"], 16)
            return datetime.datetime.utcfromtimestamp(timestamp)
        else:
            return None

    def get_logs(self, param: GetLogsParam):
        if param.toBlock:
            param.toBlock = hex(param.toBlock)
        if param.fromBlock:
            param.fromBlock = hex(param.fromBlock)
        response = self.do_post(EthRequestClient.__encode_json_rpc("eth_getLogs", [vars(param)]))
        return EthRequestClient.__decode_json_rpc(response)
